# -*- coding: utf-8 -*-

from __future__ import division


def readstr(afile):
  '''
  String data are encoded as a sequence of one or more length bytes followed
  by the specified number of data bytes.
  The lower 7 bits of each length byte encodes the bits that comprise the
  length of the following byte string.  When the most significant bit it
  set, then an additional length byte follows with 7 additional high bits to
  be added to the current length.  The following string lengths are
  accommodated by increasing sequences of length bytes:
  length  maximum
  bytes   length
  ------  --------
    1       127 B
    2        16 KB
    3         2 MB
    4       256 MB
    5        32 GB
  While this seems like a sensible progression, there is some uncertainty
  about this interpretation, since the longest of string observed in the
  wild has been of length 6,264 with two length bytes.
  '''
  cdef char *s
  cdef int n, m
  cdef bytes l

  read = afile.read

  # Read 1 byte, keeping a reference to the Python string
  # long enough to get the value
  l = read(1)
  s = l
  m = <unsigned char>s[0]

  # Fast-path return for zero length strings
  # (these occur frequently)
  if not m:
    return ''

  # Set lower 0-6 bits
  n = m&0x7F

  # Hand-unrolled code to process length bytes 2-5, when present

  # Read second length byte, if present
  if m&0x80:
    # Read 1 byte, keeping a reference to the Python string
    # long enough to get the value
    l = read(1)
    s = l
    m = <unsigned char>s[0]

    # Set bits 7-13
    n += (m&0x7F)<<7

  # Read third length byte, if present
  if m&0x80:
    # Read 1 byte, keeping a reference to the Python string
    # long enough to get the value
    l = read(1)
    s = l
    m = <unsigned char>s[0]

    # Set bits 14-20
    n += (m&0x7F)<<14

  # Read forth length byte, if present
  if m&0x80:
    # Read 1 byte, keeping a reference to the Python string
    # long enough to get the value
    l = read(1)
    s = l
    m = <unsigned char>s[0]

    # Set bits 21-27
    n += (m&0x7F)<<21

  # Read fifth length byte, if present
  if m&0x80:
    # Read 1 byte, keeping a reference to the Python string
    # long enough to get the value
    l  = read(1)
    s  = l
    m  = <unsigned char>s[0]

    # Set bits 27-33
    n += (m&0x7F)<<28

  if m&0x80:
    raise ValueError('Encoded string too long')

  return read(n)


"""Init for ReadSam."""
from __future__ import absolute_import

from .sequencing_core import process_sequencing
from .seq_utilities import read_methyldackel_bed
from .read_methyldackel import read_methyldackel


import h5py
import numpy as np
import pandas as pd
from intervalframe import IntervalFrame
import os
import glob
from typing import List, Dict, Sequence, Any
from ngsfragments.correct.correction import correct_counts

# Local imports
from .read_IDAT import read_idat, get_idat_type
from ...data.import_data import get_data_file
from ...data.download_data import download_methyl_anno
from .illumina_normalize import control_illumina_single, bgcorrect_illumina
from ..utilities import cpg_to_position, convert_v2_to_v1


def sort_chrom_names(chrom_names: Sequence[str]):
    """
    """

    # Sort chromosomes
    has_chr = chrom_names[0].startswith("chr")
    chroms = [l.split("chr")[-1] if l.startswith("chr") else l for l in chrom_names]
    int_chroms = []
    str_chroms = []
    for chrom in chroms:
        try:
            int_chroms.append(int(chrom))
        except ValueError:
            str_chroms.append(chrom)
    int_chroms.sort()
    str_chroms.sort()
    if has_chr:
        sorted_chroms = np.array(["chr"+str(c) for c in int_chroms + str_chroms])
    else:
        sorted_chroms = np.array([str(c) for c in int_chroms + str_chroms])

    return sorted_chroms


def parse_csv(filename: str):
    """
    Parse CSV formatted file
    """
    
    record = {}
    
    # iterate over lines in file
    for i, line in enumerate(open(filename,"r")):
        fields = line.strip().split(",")
        
        # if header
        if i == 0:
            n_fields = len(fields)
            header = fields
            for field in fields:
                record[field] = []  
              
        # if not header
        else:
            for j in range(n_fields):
                record[header[j]].append(fields[j])
                
    return record


def find_files(filenames: List, 
               match: Sequence[str] = None):
    # Pair up files
    file_pairs = {}
    idat_ids = set(["_".join(os.path.basename(f).split("_")[:2]) for f in filenames])
    for id in idat_ids:
        file_pairs[id] = ["tmp","tmp"]
        for f in filenames:
            if id in f:
                if "Red" in f:
                    file_pairs[id][0] = f
                else:
                    file_pairs[id][1] = f
                    
    if match is not None:
        filtered_pairs = {}
        for name in match:
            filtered_pairs[name] = file_pairs[name]

        return filtered_pairs

    else:
        return file_pairs
    

def find_files_v2(filenames: List, 
               match: Sequence[str] = None):
    

    # Pair up files
    array_types = ["27k","450k","EPIC_v1","EPIC_v2"]
    file_pairs = {t:{} for t in np.unique(array_types)}

    # Iterate over files
    #idat_ids = set(["_".join(os.path.basename(f).split("_")[:2]) for f in filenames])
    idat_ids = set(os.path.basename(f).split("_Grn")[0].split("_Red")[0] for f in filenames)
    for id in idat_ids:
        #file_pairs[id] = ["tmp","tmp"]
        for f in filenames:
            if id in f:
                array_type = get_idat_type(f)

                try:
                    file_pairs[array_type][id]
                except KeyError:
                    file_pairs[array_type][id] = ["tmp","tmp"]

                if "Red" in f:
                    file_pairs[array_type][id][0] = f
                else:
                    file_pairs[array_type][id][1] = f
                    
    if match is not None:
        filtered_pairs = {t:{} for t in np.unique(array_types)}
        for t in np.unique(array_types):
            for name in match:
                filtered_pairs[t][name] = file_pairs[t][name]

        return filtered_pairs

    else:
        return file_pairs


def find_files_v3(filenames: List, 
               match: Sequence[str] = None):
    

    # Define supported array types
    array_types = ["27k", "450k", "EPIC_v1", "EPIC_v2"]
    file_pairs = {t: {} for t in array_types}

    # Iterate through filenames to extract sample IDs and organize by array type
    for f in filenames:
        # Extract sample ID
        sample_id = os.path.basename(f).split("_Grn")[0].split("_Red")[0]
        array_type = get_idat_type(f)

        # Initialize if not present
        file_pairs[array_type].setdefault(sample_id, ["tmp", "tmp"])

        # Identify and store the correct channel
        if "Red" in f:
            file_pairs[array_type][sample_id][0] = f
        elif "Grn" in f:
            file_pairs[array_type][sample_id][1] = f

    # Filter results if 'match' is provided
    if match is not None:
        filtered_pairs = {t: {} for t in array_types}
        for t in array_types:
            for name in match:
                if name in file_pairs[t]:
                    filtered_pairs[t][name] = file_pairs[t][name]
        return filtered_pairs

    return file_pairs
        

def validate_idat_pairs(file_pairs: Dict[str, Dict]):
    """
    Remove pairs with only one idat
    """
    
    for array_type in file_pairs:
        names = list(file_pairs[array_type].keys())
        for idat_name in names:
            if file_pairs[array_type][idat_name][0] == "tmp":
                del file_pairs[array_type][idat_name]
            elif file_pairs[array_type][idat_name][1] == "tmp":
                del file_pairs[array_type][idat_name]
                
    return file_pairs


def validate_idat_pairs_v2(file_pairs: Dict[str, Dict]):
    """
    Remove pairs with only one idat.
    """

    filtered_pairs =  {array_type: {
                                idat_name: pair
                                for idat_name, pair in file_pairs[array_type].items()
                                if pair[0] != "tmp" and pair[1] != "tmp"
                                    } for array_type in file_pairs}

    return filtered_pairs


def read_all_idats(filenames: List[str],
                   verbose: bool = False):
    """
    Iteratively read idats and merge info
    """

    # Pair up files
    file_pairs = find_files(filenames)
    array_type = get_idat_type(filenames[0])
    
    # record all required idat info
    names = np.array(list(file_pairs.keys()))

    # Read all possible probes
    #man = get_data_file("hg19_manifest.h5")
    #f = h5py.File(man, "r")
    probe_files = {"27k": "./classifier_data/methyl_data_v2/microarray_data/methyl_data_v2/microarray_data/HM450_probes.parquet",
                    "450k": "./classifier_data/methyl_data_v2/microarray_data/HM450_probes.parquet",
                    "EPIC_v1": "./classifier_data/methyl_data_v2/microarray_data/EPIC_probes.parquet",
                    "EPIC_v2": "./classifier_data/methyl_data_v2/microarray_data/EPIC2_probes.parquet"}
    control_files = {"27k": "./classifier_data/methyl_data_v2/microarray_data/controls.parquet",
                        "450k": "./classifier_data/methyl_data_v2/microarray_data/controls.parquet",
                        "EPIC_v1": "./classifier_data/methyl_data_v2/microarray_data/controls.parquet",
                        "EPIC_v2": "./classifier_data/methyl_data_v2/microarray_data/EPIC2_controls.parquet"}
    snp_files = {"27k": "./classifier_data/methyl_data_v2/microarray_data/snps.parquet",
                    "450k": "./classifier_data/methyl_data_v2/microarray_data/snps.parquet",
                    "EPIC_v1": "./classifier_data/methyl_data_v2/microarray_data/snps.parquet",
                    "EPIC_v2": "./classifier_data/methyl_data_v2/microarray_data/EPIC2_snps.parquet"}
    probes_file = probe_files[array_type]
    controls_file = control_files[array_type]
    snps_file = snp_files[array_type]

    # Read control and snp probes
    #total_ids = pd.Series(np.array(f["probes"]["names"]).astype(str),
    #                        index = np.array(f["probes"]["illumina_id"]))
    #controls = pd.Series(np.array(f["controls"]["illumina_id"]),
    #                        index=np.array(f["controls"]["Type"]).astype(str))
    #snp_probes = pd.Series(np.array(f["snps"]["names"]).astype(str),
    #                        index = np.array(f["snps"]["illumina_id"]))
    #snp_probe_type = pd.Series(np.array(f["snps"]["type"]).astype(str),
    #                        index = np.array(f["snps"]["illumina_id"]))
    #probe_type = pd.Series(np.array(f["probes"]["type"]).astype(str),
    #                        index = np.array(f["probes"]["illumina_id"]))
    total_ids = pd.read_parquet(probes_file, columns=["names"]).names
    controls = pd.read_parquet(controls_file).illumina_id
    snp_probes = pd.read_parquet(snps_file, columns=["names"]).names
    snp_probe_type = pd.read_parquet(snps_file, columns=["type"]).type
    probe_type = pd.read_parquet(probes_file, columns=["type"]).type

    probe_names = pd.unique(total_ids.values)
    snp_names = pd.unique(snp_probes.values)

    #f.close()

    # Initialize DataFrames for green and red intensity values
    meth = pd.DataFrame(np.zeros((len(names), len(probe_names))),
                            index = names,
                            columns = probe_names)
    meth.values[:] = np.nan

    unmeth = pd.DataFrame(np.zeros((len(names), len(probe_names))),
                            index = names,
                            columns = probe_names)
    unmeth.values[:] = np.nan

    snps_unmeth = pd.DataFrame(np.zeros((len(names), len(snp_names))),
                        index = names,
                        columns = snp_names)
    snps_unmeth[:] = np.nan
    snps_meth = pd.DataFrame(np.zeros((len(names), len(snp_names))),
                        index = names,
                        columns = snp_names)
    snps_meth[:] = np.nan

    # Iterate over idats and normalize
    for k, name in enumerate(names):
        if verbose: print("Reading IDATs for sample %s" % name, flush=True)
        Grn = file_pairs[name][1]
        Red = file_pairs[name][0]

        idat_grn = read_idat(Grn)
        idat_red = read_idat(Red)

        control_illumina_single(idat_grn, idat_red, controls, ref=10000)
        bgcorrect_illumina(idat_grn, idat_red, controls)

        sample_probes = total_ids.index.intersection(idat_grn.index.values).values
        sample_snps = snp_probes.index.intersection(idat_grn.index.values).values
        #sample_probes = sample_probes.difference(sample_snps).values

        # Set meth
        selected = probe_type.loc[sample_probes].values == "B_I_Red"
        values = idat_red.loc[sample_probes,"means"][selected]
        #values = values.groupby(by=meth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).mean()
        values = values[~pd.Index(meth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).duplicated()]
        #meth.loc[name,values.index.values] = values.values
        meth.loc[name, total_ids.loc[values.index.values]] = values.values

        selected = probe_type.loc[sample_probes].values == "B_I_Grn"
        values = idat_grn.loc[sample_probes,"means"][selected]
        #values = values.groupby(by=meth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).mean()
        values = values[~pd.Index(meth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).duplicated()]
        #meth.loc[name,values.index.values] = values.values
        meth.loc[name, total_ids.loc[values.index.values]] = values.values
        
        selected = probe_type.loc[sample_probes].values == "A_II_Both"
        values = idat_grn.loc[sample_probes,"means"][selected]
        #values = values.groupby(by=meth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).mean()
        values = values[~pd.Index(meth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).duplicated()]
        #meth.loc[name,values.index.values] = values.values
        meth.loc[name, total_ids.loc[values.index.values]] = values.values

        # Set unmeth
        selected = probe_type.loc[sample_probes].values == "A_I_Red"
        values = idat_red.loc[sample_probes,"means"][selected]
        #values = values.groupby(by=unmeth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).mean()
        values = values[~pd.Index(unmeth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).duplicated()]
        #unmeth.loc[name,values.index.values] = values.values
        unmeth.loc[name, total_ids.loc[values.index.values]] = values.values
        
        selected = probe_type.loc[sample_probes].values == "A_I_Grn"
        values = idat_grn.loc[sample_probes,"means"][selected]
        #values = values.groupby(by=unmeth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).mean()
        values = values[~pd.Index(unmeth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).duplicated()]
        #unmeth.loc[name,values.index.values] = values.values
        unmeth.loc[name, total_ids.loc[values.index.values]] = values.values
        
        selected = probe_type.loc[sample_probes].values == "A_II_Both"
        values = idat_red.loc[sample_probes,"means"][selected]
        #values = values.groupby(by=unmeth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).mean()
        values = values[~pd.Index(unmeth.loc[name,total_ids.loc[sample_probes].values[selected]].index.values).duplicated()]
        #unmeth.loc[name,values.index.values] = values.values
        unmeth.loc[name, total_ids.loc[values.index.values]] = values.values

        # Set snps
        selected = snp_probe_type.loc[sample_snps].values == "A_I_Red"
        values = idat_red.loc[sample_snps,"means"][selected]
        #values = values.groupby(by=snps_unmeth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).mean()
        values = values[~pd.Index(snps_unmeth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).duplicated()]
        #snps_unmeth.loc[name,values.index.values] = values.values
        #snps_unmeth.loc[name, total_ids.loc[values.index.values]] = values.values
        snps_unmeth.loc[name, snp_probes.loc[values.index.values]] = values.values
        
        selected = snp_probe_type.loc[sample_snps].values == "A_I_Grn"
        values = idat_grn.loc[sample_snps,"means"][selected]
        #values = values.groupby(by=snps_unmeth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).mean()
        values = values[~pd.Index(snps_unmeth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).duplicated()]
        #snps_unmeth.loc[name,values.index.values] = values.values
        #snps_unmeth.loc[name, total_ids.loc[values.index.values]] = values.values
        snps_unmeth.loc[name, snp_probes.loc[values.index.values]] = values.values
        
        selected = snp_probe_type.loc[sample_snps].values == "A_II_Both"
        values = idat_red.loc[sample_snps,"means"][selected]
        #values = values.groupby(by=snps_unmeth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).mean()
        values = values[~pd.Index(snps_unmeth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).duplicated()]
        #snps_unmeth.loc[name,values.index.values] = values.values
        #snps_unmeth.loc[name, total_ids.loc[values.index.values]] = values.values
        snps_unmeth.loc[name, snp_probes.loc[values.index.values]] = values.values

        selected = snp_probe_type.loc[sample_snps].values == "B_I_Red"
        values = idat_red.loc[sample_snps,"means"][selected]
        #values = values.groupby(by=snps_meth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).mean()
        values = values[~pd.Index(snps_meth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).duplicated()]
        #snps_meth.loc[name,values.index.values] = values.values
        #snps_meth.loc[name, total_ids.loc[values.index.values]] = values.values
        snps_meth.loc[name, snp_probes.loc[values.index.values]] = values.values
        
        selected = snp_probe_type.loc[sample_snps].values == "B_I_Grn"
        values = idat_grn.loc[sample_snps,"means"][selected]
        #values = values.groupby(by=snps_meth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).mean()
        values = values[~pd.Index(snps_meth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).duplicated()]
        #snps_meth.loc[name,values.index.values] = values.values
        #snps_meth.loc[name, total_ids.loc[values.index.values]] = values.values
        snps_meth.loc[name, snp_probes.loc[values.index.values]] = values.values
        
        selected = snp_probe_type.loc[sample_snps].values == "A_II_Both"
        values = idat_grn.loc[sample_snps,"means"][selected]
        #values = values.groupby(by=snps_meth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).mean()
        values = values[~pd.Index(snps_meth.loc[name,snp_probes.loc[sample_snps].values[selected]].index.values).duplicated()]
        #snps_meth.loc[name,values.index.values] = values.values
        #snps_meth.loc[name, total_ids.loc[values.index.values]] = values.values
        snps_meth.loc[name, snp_probes.loc[values.index.values]] = values.values

        #snps.loc[name,sample_snps] = idat_grn.loc[sample_snps,"means"].values  / (idat_grn.loc[sample_snps,"means"].values + idat_red.loc[sample_snps,"means"].values)

    # Filter out NaNs
    selected = pd.isnull(meth.values).sum(axis=0) < meth.shape[0]
    meth = meth.loc[:,selected]
    unmeth = unmeth.loc[:,selected]

    # Drop snps
    meth.drop(columns=snps_meth.columns.values, inplace=True, errors="ignore")
    unmeth.drop(columns=snps_unmeth.columns.values, inplace=True, errors="ignore")
                
    return meth, unmeth, snps_meth, snps_unmeth


def combine_methylation_values(df1: pd.DataFrame,
                               df2: pd.DataFrame,
                               df1_type: str,
                               df2_type: str):
    """
    """

    # Combine
    if df1.shape[0] == 0:
        df1 = df2.copy()
        return df1
    
    # Check if EPIC version conversion needed
    if df1_type == "EPIC_v1" and df2_type == "EPIC_v2":
        df2 = convert_v2_to_v1(df2)
        df2 = df2.loc[:,~df2.columns.duplicated()]
    elif df1_type == "450k" and df2_type == "EPIC_v2":
        df2 = convert_v2_to_v1(df2)
        df2 = df2.loc[:,~df2.columns.duplicated()]
    
    # Add new columns
    new_columns = df2.columns.union(df1.columns).values
    df1 = df1.reindex(columns=new_columns)
    df2 = df2.reindex(columns=new_columns)

    # Concatenate
    values = np.zeros((df1.shape[0] + df2.shape[0], df1.shape[1]))
    values[:,:] = np.nan
    values[:df1.shape[0],:] = df1.values
    values[df1.shape[0]:,:] = df2.values
    df1 = pd.DataFrame(values,
                       index=np.concatenate([df1.index.values, df2.index.values]),
                       columns=new_columns)

    return df1


def multiprocess_read_all_idats(files: Sequence,
                                n_jobs: int = 1):
    """
    Read all idats in files using multiprocessing.

    Parameters
    ----------
        files : Sequence
            List of idat files.
        n_jobs : int
            Number of jobs to use.

    Returns
    -------
        meth : pd.DataFrame
            Methylation values.
        unmeth : pd.DataFrame
            Unmethylation values.
        snps_meth : pd.DataFrame
            Methylation values for SNPs.
        snps_unmeth : pd.DataFrame
            Unmethylation values for SNPs.
        array_types : np.ndarray
            Array types.
    """

    # Initialize
    total_meth = pd.DataFrame([])
    total_unmeth = pd.DataFrame([])
    total_snp_meth = pd.DataFrame([])
    total_snp_unmeth = pd.DataFrame([])
    arraytype_file_pairs = find_files_v3(files)
    arraytype_file_pairs = validate_idat_pairs_v2(arraytype_file_pairs)

    # Determine array types
    array_types = []
    for arraytype in arraytype_file_pairs.keys():
        if len(arraytype_file_pairs[arraytype]) > 0:
            array_types += [arraytype] * len(arraytype_file_pairs[arraytype])
    array_types = np.array(array_types)
    previous_array_type = None

    if n_jobs == 1:
        for arraytype in arraytype_file_pairs.keys():
            file_pairs = arraytype_file_pairs[arraytype]
            if len(file_pairs) == 0:
                continue
            # Read files
            #filenames = [reduce(lambda a,b:a+b, list(file_pairs.values())) for i in range(0, len(file_pairs))]
            filenames = np.array(list(file_pairs.values())).flatten()
            results = read_all_idats(filenames, verbose=True)

            # Combine results
            total_meth = combine_methylation_values(total_meth, results[0], previous_array_type, arraytype)
            total_unmeth = combine_methylation_values(total_unmeth, results[1], previous_array_type, arraytype)
            total_snp_meth = combine_methylation_values(total_snp_meth, results[2], previous_array_type, arraytype)
            total_snp_unmeth = combine_methylation_values(total_snp_unmeth, results[3], previous_array_type, arraytype)

            previous_array_type = arraytype

    else:
        from functools import reduce
        import multiprocessing as mp

        # Split files into n_jobs
        for arraytype in arraytype_file_pairs.keys():
            file_pairs = arraytype_file_pairs[arraytype]
            if len(file_pairs) == 0:
                continue
            n = int(np.ceil(len(file_pairs) / n_jobs))
            job_files = [reduce(lambda a,b:a+b, list(file_pairs.values())[i:i+n]) for i in range(0, len(file_pairs), n)]
            # Read files
            with mp.Pool(n_jobs) as pool:
                results = pool.map(read_all_idats, job_files)

            # Combine results
            meth = []
            unmeth = []
            snps_meth = []
            snps_unmeth = []
            for i in range(len(results)):
                meth.append(results[i][0])
                unmeth.append(results[i][1])
                snps_meth.append(results[i][2])
                snps_unmeth.append(results[i][3])
            meth_values = pd.concat(meth, axis=0)
            unmeth_values = pd.concat(unmeth, axis=0)
            snps_meth_values = pd.concat(snps_meth, axis=0)
            snps_unmeth_values = pd.concat(snps_unmeth, axis=0)

            # Combine results
            total_meth = combine_methylation_values(total_meth, meth_values, previous_array_type, arraytype)
            total_unmeth = combine_methylation_values(total_unmeth, unmeth_values, previous_array_type, arraytype)
            total_snp_meth = combine_methylation_values(total_snp_meth, snps_meth_values, previous_array_type, arraytype)
            total_snp_unmeth = combine_methylation_values(total_snp_unmeth, snps_unmeth_values, previous_array_type, arraytype)

            previous_array_type = arraytype
        
    return total_meth, total_unmeth, total_snp_meth, total_snp_unmeth, array_types


class RawArray(object):
    """
    """

    def __init__(self, 
                 data: Any = None,
                 genome_version: str = "hg38",
                 n_jobs: int = 1,
                 verbose: bool = False):
        """
        """

        self.genome_version = genome_version

        # Make sure genome is downloaded
        download_methyl_anno()

        if data is None:
            pass
        
        # load information from h5 file
        elif isinstance(data, str):
            if os.path.isfile(data) and data.endswith(".h5"):
                #load_RGset(filename)
                pass

            elif os.path.isdir(data):
                data = glob.glob(data+"**/*.idat", recursive=True)
                if verbose:
                    print("Found %i idat files" % len(data), flush=True)
                self.meth, self.unmeth, self.snps_meth, self.snps_unmeth, self.array_type = multiprocess_read_all_idats(data, n_jobs=n_jobs)
        
        elif isinstance(data, list):
            if verbose:
                print("Found %i idat files" % len(data), flush=True)
            self.meth, self.unmeth, self.snps_meth, self.snps_unmeth, self.array_type = multiprocess_read_all_idats(data, n_jobs=n_jobs)

    
    def filter_common(self):
        """
        """

        selected = pd.isnull(self.meth.values).sum(axis=0) == 0
        self.meth = self.meth.loc[:,selected]
        self.unmeth = self.unmeth.loc[:,selected]

        #selected = pd.isnull(self.snps.values).sum(axis=0) == self.snps.shape[0]
        #self.snps = self.snps.loc[:,selected]

        return
    

    def get_betas(self,
                  offset: int = 100,
                  return_intervals: bool = False):
        """
        Calculate beta values from methylation and unmethylation counts

        Parameters
        ----------
            offset : int
                Offset to add to methylation and unmethylation counts to avoid
                division by zero
            return_intervals : bool
                If True, return a IntervalFrame with beta values for each
                CpG site. If False, return a DataFrame with beta values for each
                sample
            
        Returns
        -------
            beta : pd.DataFrame or IntervalFrame
        """

        # Calculate beta
        beta = self.meth / (self.meth + self.unmeth + offset)

        # Calculate intervals
        if return_intervals:
            pos_cn = cpg_to_position(beta.columns.values,
                                     genome_version=self.genome_version)

            sorted_chroms = sort_chrom_names(pos_cn.index.unique_labels)
            pos_cn = pos_cn.loc[sorted_chroms,:]

            pos_cn.df = beta.T.loc[pos_cn.df.loc[:,"cpg_name"].values,:]

            return pos_cn

        else:
            return beta

    
    def get_cn(self,
               bias_correct: bool = False):
        """
        """

        cn = self.meth + self.unmeth

        pos_cn = cpg_to_position(cn.columns.values,
                                 genome_version=self.genome_version)

        sorted_chroms = sort_chrom_names(pos_cn.index.unique_labels)
        pos_cn = pos_cn.loc[sorted_chroms,:]

        pos_cn.df = cn.T.loc[pos_cn.df.loc[:,"cpg_name"].values,:]

        if bias_correct:
            # Assign genome
            if self.genome_version == "hg19":
                from hg19genome import calculate_bias
            elif self.genome_version == "hg38":
                from hg38genome import calculate_bias
            else:
                raise ValueError("Genome version not supported")

            # Initialize bias records
            bias_record = calculate_bias(pos_cn.index)
            
            # Remove blacklist
            chosen = bias_record.df.loc[:,"blacklist"].values < 0.1
            bias_record = bias_record.iloc[chosen,:]
            pos_cn = pos_cn.iloc[chosen,:]
            bias_record.drop_columns(["blacklist"])

            # Correct
            for sample in pos_cn.df.columns:
                pos_cn.df.loc[:,sample] = correct_counts(pos_cn.df.loc[:,sample].values,
                                                                        bias_record.df)

        return pos_cn


    def get_snps(self,
                 offset: int = 100):
        """
        """
        
        # Read SNP intervals
        if self.genome_version == "hg19":
            intervals = IntervalFrame.read_parquet("./classifier_data/methyl_data_v2/microarray_data/hg19_snp_anno.parquet")
        else:
            intervals = IntervalFrame.read_parquet("./classifier_data/methyl_data_v2/microarray_data/hg38_snp_anno.parquet")

        # Re-order
        snps = self.snps_meth / (self.snps_meth + self.snps_unmeth + offset)
        snps = snps.T
        found = np.in1d(intervals.df.loc[:,"cpg_name"].values, snps.index.values)
        intervals = intervals.iloc[found,:]
        snps = snps.loc[intervals.df.loc[:,"cpg_name"].values,:]
        snp_intervals = IntervalFrame(intervals=intervals.index, df=snps)

        return snp_intervals


import numpy as np
import scipy.stats as stats
import scipy.special as sp
import scipy.optimize as opt
import pandas as pd
import statsmodels.api as sm
import statsmodels.formula.api as smf


def beta_binomial_methylation(beta_values, sample_labels, weights=None):
    """
    Perform beta-binomial regression to compare methylation values between two samples using statsmodels.

    Parameters:
    - methylated_counts: List or array of methylated read counts.
    - total_counts: List or array of total read counts.
    - sample_labels: Binary labels (0 or 1) indicating sample groups.

    Returns:
    - Summary of regression results.
    """

    # Convert to pandas DataFrame
    data = pd.DataFrame({
        "proportion": beta_values,
        "sample": sample_labels
    })

    # Compute proportion of methylation
    data["proportion"] = data["methylated"] / data["total"]

    # Fit a binomial logistic regression (as an approximation to beta-binomial)
    if weights is None:
        model = smf.glm("proportion ~ sample",
                        data=data,
                        family=sm.families.Binomial()).fit()
    else:
        model = smf.glm("proportion ~ sample",
                        data=data,
                        family=sm.families.Binomial(),
                        freq_weights=weights).fit()

    return model


def beta_binomial_log_likelihood(params, methylated_counts, total_counts, sample_labels):
    """
    Compute the log-likelihood of the Beta-Binomial regression model.

    Parameters:
    - params: Model parameters (alpha, beta, intercept, beta_coef)
    - methylated_counts: Array of methylated read counts
    - total_counts: Array of total read counts
    - sample_labels: Binary group labels (0 = control, 1 = treatment)

    Returns:
    - Negative log-likelihood (to minimize in optimization)
    """
    alpha, beta, intercept, beta_coef = params

    # Compute probability (logit transformation)
    logit_p = intercept + beta_coef * sample_labels
    p = 1 / (1 + np.exp(-logit_p))  # Sigmoid function

    # Compute the Beta-Binomial log-likelihood
    log_likelihood = (
        sp.betaln(methylated_counts + alpha * p, total_counts - methylated_counts + beta * (1 - p))
        - sp.betaln(alpha * p, beta * (1 - p))
        + (sp.gammaln(total_counts + 1) - sp.gammaln(methylated_counts + 1) - sp.gammaln(total_counts - methylated_counts + 1))
    )

    return -np.sum(log_likelihood)  # Negative log-likelihood for minimization


def beta_binomial_methylation(methylated_counts, total_counts, sample_labels):
    """
    Perform Beta-Binomial regression using Maximum Likelihood Estimation (MLE) and compute p-values.

    Parameters:
    - methylated_counts: List or array of methylated read counts
    - total_counts: List or array of total read counts
    - sample_labels: Binary labels (0 = control, 1 = treatment)

    Returns:
    - Optimized parameter estimates (alpha, beta, intercept, beta_coef)
    - p-value for beta_coef (statistical significance test)
    """
    # Convert to numpy arrays
    methylated_counts = np.array(methylated_counts)
    total_counts = np.array(total_counts)
    sample_labels = np.array(sample_labels)

    # Initial guesses for parameters (alpha, beta, intercept, beta_coef)
    init_params = [1.0, 1.0, 0.0, 0.0]

    # Optimize using MLE
    result = opt.minimize(
        beta_binomial_log_likelihood, 
        init_params, 
        args=(methylated_counts, total_counts, sample_labels),
        method="L-BFGS-B",
        bounds=[(1e-5, None), (1e-5, None), (None, None), (None, None)],  # Alpha & Beta must be > 0
        options={'disp': False}
    )

    # Extract estimated parameters
    alpha, beta, intercept, beta_coef = result.x

    # Compute standard errors from inverse Hessian matrix (approximate variance)
    if result.hess_inv is not None:
        variance_cov_matrix = result.hess_inv.todense()  # Convert Hessian to dense matrix
        standard_errors = np.sqrt(np.diag(variance_cov_matrix))  # Get standard errors
    else:
        standard_errors = [np.nan] * 4  # Fallback if Hessian is unavailable

    # Compute z-score and p-value for beta_coef
    beta_se = standard_errors[3]  # Standard error of beta_coef
    if not np.isnan(beta_se):
        z_score = beta_coef / beta_se
        p_value = 2 * (1 - stats.norm.cdf(abs(z_score)))  # Two-tailed p-value
    else:
        p_value = np.nan  # If SE is not computable, return NaN

    return {
        "alpha": alpha,
        "beta": beta,
        "intercept": intercept,
        "beta_coef": beta_coef,
        "beta_se": beta_se,
        "z_score": z_score,
        "p_value": p_value,
        "log_likelihood": -result.fun  # Log-likelihood of the model
    }


"""Init for ReadSam."""
from __future__ import absolute_import
from .read_IDAT import read_idat
from .RawArray import RawArray, read_all_idats, find_files_v2, find_files_v3, multiprocess_read_all_idats, validate_idat_pairs



import numpy as np
from scipy.optimize import minimize
from scipy.special import betaln
from scipy.stats import beta, norm


def beta_mean(a, b):
    """
    Compute the mean of a Beta distribution.
    """
    
    return a / (a + b)

def beta_var(a, b):
    """
    Compute the variance of a Beta distribution.
    """

    return (a * b) / ((a + b)**2 * (a + b + 1))

def beta_std(a, b):
    """
    Compute the standard deviation of a Beta distribution.
    """

    return np.sqrt(beta_var(a, b))


def beta_median(a, b):
    """
    Compute the median of a Beta distribution.
    """

    return beta.ppf(0.5, a, b)


def update_beta_prior(alpha, beta, x):
    """
    Update the prior parameters of a Beta distribution based on observed data.
    alpha, beta: initial parameters of the Beta distribution.
    x: array of observed values (assumed to be in (0,1)).
    """
    # Count successes and failures
    successes = np.sum(x)
    failures = len(x) - successes
    # Update parameters
    new_alpha = alpha + successes
    new_beta = beta + failures
    return new_alpha, new_beta

def beta_neg_log_likelihood(params, x):
    """
    Compute the negative log likelihood for the Beta distribution.
    x: array of methylation values (assumed to be in (0,1)).
    params: tuple (a, b) for the Beta distribution parameters.
    """
    a, b = params
    # enforce a, b > 0
    if a <= 0 or b <= 0:
        return np.inf
    n = len(x)
    # The log likelihood for a Beta distribution is:
    #   sum[(a-1)*log(x_i) + (b-1)*log(1-x_i)] - n * betaln(a, b)
    ll = np.sum((a - 1) * np.log(x) + (b - 1) * np.log(1 - x)) - n * betaln(a, b)
    return -ll  # return negative log-likelihood for minimization

def estimate_beta_params(x, init_params=(2.0, 2.0)):
    """
    Estimate Beta distribution parameters from data using maximum likelihood.
    x: array of methylation values (must be in (0,1); values are clipped to avoid 0,1 issues).
    init_params: initial guess for (a, b).
    """
    # Clip x slightly to avoid issues with log(0)
    x = np.clip(x, 1e-6, 1 - 1e-6)
    result = minimize(beta_neg_log_likelihood, x0=init_params, args=(x,),
                      bounds=[(1e-3, None), (1e-3, None)])
    if result.success:
        return result.x  # returns (a, b)
    else:
        raise RuntimeError("Beta parameter estimation failed.")

def log_likelihood_beta(x, a, b):
    """
    Compute the log likelihood of data x under a Beta(a, b) model.
    """
    x = np.clip(x, 1e-6, 1 - 1e-6)
    return np.sum((a - 1) * np.log(x) + (b - 1) * np.log(1 - x) - betaln(a, b))
    
    
def beta_model_test(x, alpha=2, beta=1):
    try:
        # Estimate Beta parameters (MLE) for this region
        a_mle, b_mle = estimate_beta_params(x)
    except RuntimeError:
        print("Parameter estimation failed.")
        return None, None
    # Compute the estimated mean methylation level for the region
    mle_mean = a_mle / (a_mle + b_mle)
    
    # Compute log likelihoods under the estimated model and under the fixed high methylation model
    ll_mle = log_likelihood_beta(x, a_mle, b_mle)
    ll_high = log_likelihood_beta(x, alpha, beta)
    
    # Log-likelihood ratio (a positive value indicates the estimated model fits better than the high methylation model)
    llr = ll_mle - ll_high
    
    return mle_mean, llr
    
def beta_zscore(x_obs, alpha=2, beta_param=1):
    """
    Compute a standardized score for observed methylation values under a Beta distribution.
    x_obs: observed methylation value (assumed to be in (0,1)).
    alpha, beta_param: parameters of the Beta distribution.
    """
    # Compute mean and standard deviation of the Beta distribution
    mean_beta = alpha / (alpha + beta_param)
    std_beta = np.sqrt((alpha * beta_param) / ((alpha + beta_param)**2 * (alpha + beta_param + 1)))

    # Compute "Beta z-score" (standardized beta score)
    z_beta = (x_obs - mean_beta) / std_beta

    # Compute percentile-based score (p-value equivalent)
    p_value = beta.cdf(x_obs, alpha, beta_param)

    # Convert percentile to standard normal z-score equivalent
    z_p = norm.ppf(p_value)
    
    return z_beta, z_p, p_value


def beta_tscore(x, y):
    """
    """

    # Fit Beta distribution to data
    x_alpha, x_beta = estimate_beta_params(x)
    y_alpha, y_beta = estimate_beta_params(y)

    # Compute mean and standard deviation of the Beta distribution
    x_mean = beta_mean(x_alpha, x_beta)
    x_std = beta_std(x_alpha, x_beta)
    y_mean = beta_mean(y_alpha, y_beta)
    y_std = beta_std(y_alpha, y_beta)

    # Compute "Beta z-score" (standardized beta score)
    t_beta = (x_mean - y_mean) / (x_std / np.sqrt(len(x)))

    return t_beta
    


import numpy as np
import pandas as pd
from typing import List
import os
import ngsfragments as ngs

# Local imports
from ..data.import_data import get_data_file
from ..data.download_data import download_MPACT
from ..tools.classifiers.MPACT.MPACT_classifier_torch import MPACT_classifier
from ..tools.decomposition.reference_decom import remove_normal_csf, tumor_decomposition
from ..tools.imputation.imputeEPIC import impute_epic, EPICimputer
from ..tools.cnvs.seq_cnv_calling import sequencing_call_cnvs
from ..tools.cnvs.idat_cnv_calling import microarray_call_cnvs
from ..core.methyl_core import read_sequencing_methylation, RawArray
from ..core.utilities import position_to_cpg


def MPACT_process_raw(input_data: List[str],
                        impute: bool = False,
                        regress: bool = False,
                        probability_threshold: float = 0.7,
                        max_contamination_fraction: float = 0.5,
                        call_cnvs: bool = False,
                        output_cnv: str = None,
                        verbose: bool = False) -> pd.DataFrame:
    """
    """
    # Check data downloaded
    #download_MPACT()

    # Initialize
    is_sequencing = True
    if len(input_data) > 1:
        is_sequencing = False

    # Read methylation
    if verbose: print("Reading methylation data", flush=True)
    if is_sequencing:
        sample_name = os.path.basename(input_data[0]).split(".")[0]
        data = read_sequencing_methylation(input_data[0])
        cpgs = position_to_cpg(data, genome_version="hg38")
        data = data.df.loc[cpgs!='', :]
        data.index = cpgs[cpgs!='']
        data = data.T
        data.index = [sample_name]
        data = data.loc[:,~data.columns.duplicated()]
    else:
        if "Red" in input_data[0]:
            sample_name = os.path.basename(input_data[0]).split("_Red")[0]
        else:
            sample_name = os.path.basename(input_data[0]).split("_Grn")[0]
        raw_array = RawArray(input_data, genome_version="hg38", n_jobs=1, verbose=False)
        data = raw_array.get_betas()
        data.index = [sample_name]

    # Impute
    if verbose: print("Imputing data", flush=True)
    if impute:
        data = impute_epic(data, method="regress")
    else:
        imputer = EPICimputer()
        # Find missing features
        missing = imputer.features[~pd.Series(imputer.features).isin(data.columns)]
        # Set missing features to nan
        null_df = pd.DataFrame(np.nan, index=data.index, columns=missing)
        data = pd.concat([data, null_df], axis=1)
        data = data.loc[:,imputer.features]

    # Regress
    if regress:
        if verbose: print("Regressing data", flush=True)
        transformed_data = remove_normal_csf(data, max_fraction=max_contamination_fraction)

    # Classify
    # Get classifier files
    if verbose: print("Classifying data", flush=True)
    #onnx_directory = os.path.dirname(get_data_file("MPACT_D1_v1.pth"))

    classifier = MPACT_classifier("/opt/data/")
    if regress:
        transformed_data = (transformed_data * 2) - 1
    else:
        transformed_data = (data * 2) - 1
    transformed_data.fillna(0, inplace=True)
    predictions = classifier.predict(transformed_data, transform=False)
    prediction = predictions[0][0]
    probability = predictions[1].loc[:,prediction].values[0]

    # Deconvolve
    if verbose: print("Deconvolving data", flush=True)
    if probability > probability_threshold and prediction != "NonmalignantBackground":
        decon = tumor_decomposition(data, [prediction], verbose=False)
        purity = decon.values[0]
        if purity == 0 and prediction != "NonmalignantBackground":
            prediction = "NonmalignantBackground"
            purity = 1.0
            probability = 1.0
    elif probability < probability_threshold and prediction == "NonmalignantBackground":
        secondary_prediction = predictions[1].columns.values[predictions[1].values.argsort()[0][-2]]
        decon = tumor_decomposition(data, [secondary_prediction], verbose=False)
        # GCT corresponds to over regressed samples meaning no tumor
        if decon.values[0] == 0 or secondary_prediction == "GCT":
            purity = 0.0
            probability = 1.0
    else:
        purity = 0.0
    
    # Call CNVs
    if call_cnvs:
        if verbose: print("Calling CNVs", flush=True)
        if is_sequencing:
            data = read_sequencing_methylation(input_data[0], read_cov=True)
            data.df.columns = [sample_name]
            cnvs = sequencing_call_cnvs(data)

            # Plot CNVs
            if output_cnv is not None:
                ngs.plot.plot_cnv(cnvs.pf,
                                obs = sample_name,
                                show = False,
                                save = output_cnv,
                                plot_max = 5,
                                plot_min = -3)
            else:
                ngs.plot.plot_cnv(cnvs.pf,
                                    obs = sample_name,
                                    show = False,
                                    save = sample_name+"_cnvs.pdf",
                                    plot_max = 5,
                                    plot_min = -3)
        else:
            cnvs = microarray_call_cnvs(raw_array)

            # Plot CNVs
            if output_cnv is not None:
                ngs.plot.plot_cnv(cnvs.pf,
                                obs = sample_name,
                                show = False,
                                save = output_cnv,
                                plot_max = 2,
                                plot_min = -2)
            else:
                ngs.plot.plot_cnv(cnvs.pf,
                                    obs = sample_name,
                                    show = False,
                                    save = sample_name+"_cnvs.pdf",
                                    plot_max = 2,
                                    plot_min = -2)
        if verbose: print("Output plots saved:", sample_name+"_cnvs.pdf", flush=True)

        # Construct results
        if verbose: print("Constructing results", flush=True)
        cnv_purity = cnvs.pf.anno.loc[sample_name, "purity"]
        cnv_ploidy = cnvs.pf.anno.loc[sample_name, "ploidy"]
        results = pd.DataFrame([prediction, probability, purity, cnv_purity, cnv_ploidy], columns = [sample_name],
                            index=["MPACT_Prediction", "MPACT_Probability", "MPACT_Purity", "CNV_Purity", "CNV_Ploidy"]).T
    else:
        results = pd.DataFrame([prediction, probability, purity], columns = [sample_name],
                            index=["MPACT_Prediction", "MPACT_Probability", "MPACT_Purity"]).T
    
    return results


import numpy as np
import pandas as pd
from scipy.optimize import minimize
from scipy.special import betaln
from scipy.stats import beta, norm, combine_pvalues
from scipy.special import logit
import statsmodels.api as sm
from statsmodels.formula.api import glm
from statsmodels.stats.multitest import multipletests
from sklearn.gaussian_process import GaussianProcessRegressor
from sklearn.gaussian_process.kernels import RBF, WhiteKernel
from ailist import LabeledIntervalArray
from intervalframe import IntervalFrame
from scipy.interpolate import UnivariateSpline
from sklearn.preprocessing import StandardScaler
from scipy import stats
import statsmodels.formula.api as smf


def beta_mean(a, b):
    """
    Compute the mean of a Beta distribution.
    """
    return a / (a + b)


def beta_var(a, b):
    """
    Compute the variance of a Beta distribution.
    """
    return (a * b) / ((a + b)**2 * (a + b + 1))


def beta_std(a, b):
    """
    Compute the standard deviation of a Beta distribution.
    """
    return np.sqrt(beta_var(a, b))


def beta_median(a, b):
    """
    Compute the median of a Beta distribution.
    """
    return beta.ppf(0.5, a, b)


def update_beta_prior(alpha, beta, x):
    """
    Update the prior parameters of a Beta distribution based on observed data.
    alpha, beta: initial parameters of the Beta distribution.
    x: array of observed values (assumed to be in (0,1)).
    """
    # Count successes and failures
    successes = np.sum(x)
    failures = len(x) - successes
    # Update parameters
    new_alpha = alpha + successes
    new_beta = beta + failures
    return new_alpha, new_beta


def beta_neg_log_likelihood(params, x):
    """
    Compute the negative log likelihood for the Beta distribution.
    x: array of methylation values (assumed to be in (0,1)).
    params: tuple (a, b) for the Beta distribution parameters.
    """
    a, b = params
    # enforce a, b > 0
    if a <= 0 or b <= 0:
        return np.inf
    n = len(x)
    # The log likelihood for a Beta distribution
    ll = np.sum((a - 1) * np.log(x) + (b - 1) * np.log(1 - x)) - n * betaln(a, b)
    return -ll  # return negative log-likelihood for minimization


def estimate_beta_params(x, init_params=(2.0, 2.0)):
    """
    Estimate Beta distribution parameters from data using maximum likelihood.
    x: array of methylation values (must be in (0,1); values are clipped to avoid 0,1 issues).
    init_params: initial guess for (a, b).
    """
    # Handle edge cases where all values are close to 0 or 1
    if np.all(x <= 0) or np.all(x >= 1):
        raise ValueError("All values must be within the open interval (0, 1).")

    # Clip x slightly to avoid issues with log(0)
    x = np.clip(x, 1e-6, 1 - 1e-6)
    result = minimize(beta_neg_log_likelihood, x0=init_params, args=(x,),
                      bounds=[(1e-3, None), (1e-3, None)])
    if result.success:
        return result.x  # returns (a, b)
    else:
        raise RuntimeError("Beta parameter estimation failed.")


def log_likelihood_beta(x, a, b):
    """
    Compute the log likelihood of data x under a Beta(a, b) model.
    """
    x = np.clip(x, 1e-6, 1 - 1e-6)
    return np.sum((a - 1) * np.log(x) + (b - 1) * np.log(1 - x) - betaln(a, b))


def beta_model_test(x, alpha=2, beta=1):
    """
    Compare fitted Beta model to a reference high methylation Beta model.

    Parameters:
    - x: array of methylation values (assumed to be in (0,1)).
    - alpha, beta: parameters of the reference Beta distribution.

    Returns:
    - mle_mean: Mean of the estimated Beta distribution.
    - llr: Log-likelihood ratio comparing fitted model to reference.
    """
    try:
        # Estimate Beta parameters (MLE) for this region
        a_mle, b_mle = estimate_beta_params(x)
    except (RuntimeError, ValueError):
        print("Parameter estimation failed.")
        return None, None

    # Compute the estimated mean methylation level for the region
    mle_mean = a_mle / (a_mle + b_mle)

    # Compute log likelihoods under the estimated model and under the fixed high methylation model
    ll_mle = log_likelihood_beta(x, a_mle, b_mle)
    ll_high = log_likelihood_beta(x, alpha, beta)

    # Log-likelihood ratio
    llr = ll_mle - ll_high

    return mle_mean, llr


def beta_zscore(x_obs, alpha=2, beta_param=1):
    """
    Compute a standardized score for observed methylation values under a Beta distribution.
    """
    mean_beta = beta_mean(alpha, beta_param)
    std_beta = beta_std(alpha, beta_param)

    # Compute "Beta z-score"
    z_beta = (x_obs - mean_beta) / std_beta

    # Compute percentile-based score (p-value equivalent)
    p_value = beta.cdf(x_obs, alpha, beta_param)

    # Convert percentile to standard normal z-score equivalent
    z_p = norm.ppf(p_value)

    return z_beta, z_p, p_value


def beta_tscore(x, y):
    """
    Compute a t-like statistic comparing the means of two Beta-distributed datasets
    using the Delta method for more accurate variance estimation.

    Parameters:
    - x: First dataset (array of methylation values in (0, 1)).
    - y: Second dataset (array of methylation values in (0, 1)).

    Returns:
    - t_beta: The t-like statistic comparing the means of the two datasets.
    - p_value: Two-tailed p-value for the difference in means.
    """

    # Remove NaNs
    x = x[~np.isnan(x)]
    y = y[~np.isnan(y)]

    # Fit Beta distribution to data
    x_alpha, x_beta = estimate_beta_params(x)
    y_alpha, y_beta = estimate_beta_params(y)

    # Means of the Beta distributions
    x_mean = beta_mean(x_alpha, x_beta)
    y_mean = beta_mean(y_alpha, y_beta)

    # Variances using the delta method
    x_var = (x_alpha * x_beta) / ((x_alpha + x_beta) ** 2 * (x_alpha + x_beta + 1))
    y_var = (y_alpha * y_beta) / ((y_alpha + y_beta) ** 2 * (y_alpha + y_beta + 1))

    # Standard error of the difference in means
    se_diff = np.sqrt(x_var / len(x) + y_var / len(y))

    # T-like statistic
    t_beta = (x_mean - y_mean) / se_diff

    # Compute two-tailed p-value using normal approximation
    p_value = 2 * (1 - norm.cdf(np.abs(t_beta)))

    return t_beta, p_value


def calc_autocorrelation(methylation_matrix, cpg_positions, max_lag=500):
    """
    Calculate spatial autocorrelation between CpG sites.

    Parameters:
    - methylation_matrix: 2D numpy array (CpGs as rows, samples as columns) with methylation values (0-1 range).
    - cpg_positions: 1D array of genomic positions corresponding to each CpG row.
    - max_lag: Maximum genomic distance to consider for autocorrelation.

    Returns:
    - autocorrelations: List of autocorrelation values for each lag distance.
    """
    # Ensure correct input shapes
    assert methylation_matrix.shape[0] == len(cpg_positions), "Mismatch between rows and CpG positions."

    # Number of CpG sites
    n_cpgs, n_samples = methylation_matrix.shape

    # Mean-center the methylation values across samples
    centered_methylation = methylation_matrix - np.mean(methylation_matrix, axis=1, keepdims=True)

    autocorrelations = []

    # Loop over possible lags (spatial distances between CpGs)
    for lag in range(1, max_lag + 1):
        # Track correlation at this lag
        valid_pairs = []
        correlations = []

        # Compare each CpG with its neighbors within the lag range
        for i in range(n_cpgs - lag):
            # Find the next CpG within the lag
            j = i + 1
            while j < n_cpgs and (cpg_positions[j] - cpg_positions[i]) <= lag:
                # Calculate Pearson correlation between methylation levels
                corr = np.corrcoef(centered_methylation[i], centered_methylation[j])[0, 1]
                correlations.append(corr)
                valid_pairs.append((i, j))
                j += 1

        # Average correlation for this lag
        if correlations:
            autocorrelations.append(np.mean(correlations))
        else:
            autocorrelations.append(np.nan)  # No valid pairs for this lag

    return np.array(autocorrelations)


def beta_regression_pvalues(methylation_matrix, group_labels):
    """
    Perform Beta regression for each CpG site to calculate p-values for group differences.

    Parameters:
    - methylation_matrix: 2D numpy array (CpGs as rows, samples as columns) with methylation values in (0, 1).
    - group_labels: 1D array-like with sample group labels (categorical).

    Returns:
    - p_values: Array of p-values for each CpG site.
    """
    n_cpgs, n_samples = methylation_matrix.shape
    group_labels = np.array(group_labels)

    # Ensure the matrix and labels are aligned
    assert len(group_labels) == n_samples, "Mismatch between samples and labels."
    unique_groups = pd.unique(group_labels)

    # Preprocess methylation values (clip to avoid log(0) issues)
    methylation_matrix = np.clip(methylation_matrix, 1e-6, 1 - 1e-6)

    p_values = []

    for cpg_idx in range(n_cpgs):
        # Extract methylation values for this CpG site
        methylation_values = methylation_matrix[cpg_idx, :]

        # Prepare the data for regression
        df = pd.DataFrame({
            'methylation': methylation_values,
            'group': group_labels
        })

        # Apply logit transformation to map (0, 1) → (-inf, inf)
        df['methylation_logit'] = logit(df['methylation'])

        # Fit a Beta regression model (using logit link function for the mean)
        model = glm("methylation_logit ~ C(group)", data=df,
                    family=sm.families.Binomial()).fit()

        # Extract p-value for the group effect
        p_value = model.pvalues["C(group)[T." + str(unique_groups[1]) + "]"]
        p_values.append(p_value)

    return np.array(p_values)


def smooth_pvalues_rolling(pvalues, positions, max_dist=500):
    """
    Smooth p-values using spatial correlation between CpG sites.

    Parameters:
    - pvalues: Array of raw p-values for each CpG site.
    - positions: Array of genomic positions for each CpG site.
    - max_dist: Maximum genomic distance to consider for smoothing.

    Returns:
    - smoothed_pvalues: Smoothed p-values accounting for spatial correlation.
    """
    n = len(pvalues)
    smoothed_pvalues = np.zeros(n)

    for i in range(n):
        # Identify neighboring CpGs within the specified distance
        neighbors = np.where(np.abs(positions - positions[i]) <= max_dist)[0]

        # Use Fisher’s method to combine p-values of neighboring CpGs
        combined_pval = combine_pvalues(pvalues[neighbors], method='fisher')[1]
        smoothed_pvalues[i] = combined_pval

    return smoothed_pvalues


def smooth_pvalues_window(pvalues, positions, window_size=1000):
    z_scores = -norm.ppf(np.clip(pvalues, 1e-10, 1 - 1e-10))
    smoothed_z = np.zeros_like(z_scores)
    
    # Sort by position
    idx = np.argsort(positions)
    sorted_pos = positions[idx]
    sorted_z = z_scores[idx]
    
    # Sliding window average
    for i in range(len(sorted_pos)):
        window = np.abs(sorted_pos - sorted_pos[i]) <= window_size
        smoothed_z[idx[i]] = np.mean(sorted_z[window])
    
    # Convert back to p-values
    return 2 * (1 - norm.cdf(np.abs(smoothed_z)))


def smooth_pvalues_spline(pvalues, positions, smoothing=0.5):
    z_scores = -norm.ppf(np.clip(pvalues, 1e-10, 1 - 1e-10))
    
    # Sort by position
    idx = np.argsort(positions)
    sorted_pos = positions[idx]
    sorted_z = z_scores[idx]
    
    # Fit spline (s controls smoothing strength)
    spline = UnivariateSpline(sorted_pos, sorted_z, s=smoothing*len(sorted_pos))
    smoothed_z = spline(positions)
    
    # Convert back to p-values
    return 2 * (1 - norm.cdf(np.abs(smoothed_z)))


def correct_pvalues_autocorr(pvalues, positions, max_dist=500, method='fdr_bh'):
    """
    Perform p-value correction accounting for spatial correlation.

    Parameters:
    - pvalues: Array of raw p-values for each CpG site.
    - positions: Array of genomic positions for each CpG site.
    - max_dist: Maximum genomic distance for smoothing.
    - method: Multiple testing correction method (default: 'fdr_bh').

    Returns:
    - adjusted_pvalues: P-values adjusted for multiple testing.
    """
    # Step 1: Smooth p-values using spatial autocorrelation
    smoothed_pvalues = smooth_pvalues(pvalues, positions, max_dist)

    # Step 2: Apply multiple testing correction (Benjamini-Hochberg or others)
    _, adjusted_pvalues, _, _ = multipletests(smoothed_pvalues, method=method)

    return adjusted_pvalues


def smooth_pvalues_gp(pvalues, positions, length_scale=1000.0, noise_level=1e-4):
    positions = np.array(positions).reshape(-1, 1)

    # Transform p-values to z-scores
    z_scores = -norm.ppf(np.clip(pvalues, 1e-10, 1 - 1e-10))

    # Gaussian Process with RBF kernel (for smoothing) and noise (to prevent overfitting)
    kernel = RBF(length_scale=length_scale) + WhiteKernel(noise_level=noise_level)
    gp = GaussianProcessRegressor(kernel=kernel, normalize_y=True)

    # Fit GP to z-scores
    gp.fit(positions, z_scores)

    # Predict smoothed z-scores
    smoothed_z = gp.predict(positions)

    # Convert back to p-values
    smoothed_pvalues = 2 * (1 - norm.cdf(np.abs(smoothed_z)))

    return smoothed_pvalues

def identify_clusters(positions, pvalues, threshold=0.05, max_gap=500):
    significant_indices = np.where(pvalues < threshold)[0]
    clusters = []
    current_cluster = [significant_indices[0]]

    for i in range(1, len(significant_indices)):
        if positions[significant_indices[i]] - positions[significant_indices[i - 1]] <= max_gap:
            current_cluster.append(significant_indices[i])
        else:
            clusters.append(current_cluster)
            current_cluster = [significant_indices[i]]

    if current_cluster:
        clusters.append(current_cluster)

    return clusters

def combine_cluster_pvalues(pvalues, clusters):
    combined_pvalues = []

    for cluster in clusters:
        cluster_pvalues = pvalues[cluster]
        if len(cluster_pvalues) == 1:
            combined_pvalues.append(cluster_pvalues[0])
        else:
            stat, combined_p = combine_pvalues(cluster_pvalues, method='fisher')
            combined_pvalues.append(combined_p)

    return np.array(combined_pvalues)

def call_dmps(data, group_labels, method='ttest', 
             covariates=None, alpha=0.05, correction='fdr_bh', 
             logit_transform=True, min_delta=0.05):
    """
    Find differentially methylated positions (DMPs) between groups using various statistical methods.
    
    Parameters:
    -----------
   data : IntervalFrame
        2D array with methylation beta values (shape: n_cpgs × n_samples)
    group_labels : numpy.ndarray
        1D array with group assignments for each sample
    method : str
        Statistical method to use: 'ttest', 'mannwhitney', 'logistic', or 'beta'
    covariates : numpy.ndarray or pandas.DataFrame, optional
        2D array of covariates (shape: n_samples × n_covariates)
    alpha : float
        Significance level for adjusted p-values
    correction : str
        Method for multiple testing correction (see statsmodels.stats.multitest)
    logit_transform : bool
        Whether to logit transform methylation values for t-tests (recommended)
    min_delta : float
        Minimum absolute methylation difference to consider a DMP significant
        
    Returns:
    --------
    pandas.DataFrame
        Results with statistics, p-values, and significance calls
    """
    methylation_data = data.df.values
    unique_groups = np.unique(group_labels)
    if len(unique_groups) != 2:
        raise ValueError("This function currently supports only two groups")
    
    n_cpgs = methylation_data.shape[0]
    n_samples = methylation_data.shape[1]
    
    if len(group_labels) != n_samples:
        raise ValueError(f"group_labels length ({len(group_labels)}) must match the number of samples ({n_samples})")
    
    # Convert group labels to binary (0/1)
    group_binary = np.zeros(n_samples)
    group_binary[group_labels == unique_groups[1]] = 1
    
    group1_idx = group_labels == unique_groups[0]
    group2_idx = group_labels == unique_groups[1]
    
    # Initialize results storage
    pvals = []
    stats_values = []
    delta_betas = []
    coef_values = []
    
    # Process each CpG site
    for cpg_idx in range(n_cpgs):
        # Extract methylation values for this CpG
        cpg_values = methylation_data[cpg_idx, :]
        
        # Calculate mean difference between groups
        mean_group1 = np.mean(cpg_values[group1_idx])
        mean_group2 = np.mean(cpg_values[group2_idx])
        delta = mean_group2 - mean_group1
        delta_betas.append(delta)
        
        # Apply statistical test based on selected method
        if method == 'ttest':
            values_group1 = cpg_values[group1_idx]
            values_group2 = cpg_values[group2_idx]
            
            # Apply logit transform if requested (handles bounds better)
            if logit_transform:
                # Avoid log(0) and log(1) issues
                epsilon = 1e-6
                values_group1 = np.clip(values_group1, epsilon, 1-epsilon)
                values_group2 = np.clip(values_group2, epsilon, 1-epsilon)
                values_group1 = np.log(values_group1 / (1 - values_group1))
                values_group2 = np.log(values_group2 / (1 - values_group2))
            
            t_stat, p_val = stats.ttest_ind(values_group1, values_group2, equal_var=False)
            pvals.append(p_val)
            stats_values.append(t_stat)
            coef_values.append(delta)
            
        elif method == 'mannwhitney':
            values_group1 = cpg_values[group1_idx]
            values_group2 = cpg_values[group2_idx]
            u_stat, p_val = stats.mannwhitneyu(values_group1, values_group2)
            pvals.append(p_val)
            stats_values.append(u_stat)
            coef_values.append(delta)
            
        elif method == 'logistic':
            # For logistic regression, methylation is predictor, group is outcome
            X = cpg_values.reshape(-1, 1)
            y = group_binary
            
            # Add intercept and standardize methylation values
            X = StandardScaler().fit_transform(X)
            X = sm.add_constant(X)
            
            # Add covariates if provided
            if covariates is not None:
                if isinstance(covariates, np.ndarray):
                    X = np.column_stack((X, covariates))
                else:
                    X = np.column_stack((X, covariates.values))
            
            try:
                # Fit logistic regression model
                model = sm.Logit(y, X)
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    results = model.fit(disp=0, method='bfgs')
                
                # Extract coefficient and p-value for methylation effect (index 1)
                pvals.append(results.pvalues[1])
                stats_values.append(results.tvalues[1])
                coef_values.append(results.params[1])
            except:
                # Handle convergence issues
                pvals.append(1.0)
                stats_values.append(0.0)
                coef_values.append(0.0)
                
        elif method == 'beta':
            # Proper implementation for beta regression
            # Uses a linear model with logit-transformed methylation values
            
            # Avoid exact 0 and 1 values
            epsilon = 1e-6
            cpg_values_adj = np.clip(cpg_values, epsilon, 1-epsilon)
            
            # Logit transform methylation values - necessary for beta regression approximation
            logit_values = np.log(cpg_values_adj / (1 - cpg_values_adj))
            
            # Create design matrix with group and covariates
            X = sm.add_constant(group_binary.reshape(-1, 1))
            
            # Add covariates if provided
            if covariates is not None:
                if isinstance(covariates, np.ndarray):
                    X = np.column_stack((X, covariates))
                else:
                    X = np.column_stack((X, covariates.values))
            
            try:
                # Fit linear model on logit-transformed values
                model = sm.OLS(logit_values, X)
                results = model.fit()
                
                # Extract coefficient and p-value for group effect (index 1)
                pvals.append(results.pvalues[1])
                stats_values.append(results.tvalues[1])
                coef_values.append(results.params[1])
            except Exception as e:
                # Handle fitting issues
                pvals.append(1.0)
                stats_values.append(0.0)
                coef_values.append(0.0)
        else:
            raise ValueError(f"Method '{method}' not recognized. Use 'ttest', 'mannwhitney', 'logistic', or 'beta'")
    
    # Multiple testing correction
    reject, pvals_corrected, _, _ = multipletests(pvals, alpha=alpha, method=correction)
    
    # Create results DataFrame
    results = IntervalFrame(intervals=data.index,
                            df = pd.DataFrame({
                            'cpg_index': range(n_cpgs),
                            'p_value': pvals,
                            'adjusted_p_value': pvals_corrected,
                            'statistic': stats_values,
                            'effect_size': coef_values,
                            'delta_beta': delta_betas,
                            'is_significant': reject & (np.abs(np.array(delta_betas)) >= min_delta)
                        }))
    
    return results


def detect_chrom_dmrs(pvalues,
                        positions,
                        chromosome,
                        length_scale=1000,
                        noise_level=1e-4,
                        pvalue_threshold=0.05,
                        max_gap=500,
                        min_cpgs=3,
                        smooth_method="spline"):
    # Step 1: Smooth p-values using GP regression
    if smooth_method == "spline":
        smoothed_pvalues = smooth_pvalues_spline(pvalues, positions)
    elif smooth_method == "window":
        smoothed_pvalues = smooth_pvalues_window(pvalues, positions)
    elif smooth_method == "gp":
        smoothed_pvalues = smooth_pvalues_gp(pvalues, positions, length_scale, noise_level)
    elif smooth_method == "rolling":
        smoothed_pvalues = smooth_pvalues_rolling(pvalues, positions, max_gap)
    else:
        raise ValueError("Invalid smoothing method. Choose 'spline', 'window', or 'gp'.")

    # Step 2: Identify CpG clusters
    clusters = identify_clusters(positions, smoothed_pvalues, pvalue_threshold, max_gap)

    # Step 3: Filter clusters by minimum CpG count
    clusters = [c for c in clusters if len(c) >= min_cpgs]

    # Step 4: Combine p-values within each cluster
    combined_pvalues = combine_cluster_pvalues(smoothed_pvalues, clusters)

    # Step 5: Adjust cluster p-values for multiple testing
    _, adjusted_pvalues, _, _ = multipletests(combined_pvalues, method='fdr_bh')

    # Prepare DMR results
    dmr_results = pd.DataFrame({
        'cluster_id': range(len(clusters)),
        'start_position': [positions[c[0]] for c in clusters],
        'end_position': [positions[c[-1]] for c in clusters],
        'num_cpgs': [len(c) for c in clusters],
        'combined_pvalue': combined_pvalues,
        'adjusted_pvalue': adjusted_pvalues
    })

    # Create IntervalFrame for DMRs
    intervals = LabeledIntervalArray()
    intervals.add(dmr_results['start_position'].values, dmr_results['end_position'].values, np.repeat(chromosome,dmr_results.shape[0]))
    dmr_intervals = IntervalFrame(df=dmr_results, intervals=intervals)
    dmr_intervals.df = dmr_intervals.df.drop(columns=['start_position', 'end_position'])

    return dmr_intervals


def detect_dmrs(pvalues,
                length_scale=1000,
                noise_level=1e-4,
                pvalue_threshold=0.05,
                max_gap=500,
                min_cpgs=3,
                smooth_method="spline",
                verbose=False):
    """
    Detect differentially methylated regions (DMRs) based on p-values and genomic positions.
    """

    # Iterate through chromosomes
    dmr_results = []
    unique_chromosomes = pvalues.index.unique_labels
    for chrom in unique_chromosomes:
        if verbose:
            print(f"Processing chromosome: {chrom}", flush=True)

        # Extract p-values and positions for this chromosome
        chrom_pvalues = pvalues.loc[chrom, :].df.loc[:,"p_value"].values
        chrom_positions = pvalues.loc[chrom, :].index.starts

        # Detect DMRs for this chromosome
        if len(chrom_pvalues) >= 10:
            dmr_intervals = detect_chrom_dmrs(chrom_pvalues,
                                            chrom_positions,
                                            chrom,
                                            length_scale=length_scale,
                                            noise_level=noise_level,
                                            pvalue_threshold=pvalue_threshold,
                                            max_gap=max_gap,
                                            min_cpgs=min_cpgs,
                                            smooth_method=smooth_method)
            dmr_results.append(dmr_intervals)

    # Combine results from all chromosomes
    dmrs = dmr_results[0].concat(dmr_results[1:])
    dmrs.df.loc[:,"cluster_id"] = np.arange(dmrs.df.shape[0])

    return dmrs


from os.path import join
import os
import requests
import shutil
from tqdm.auto import tqdm
import json
import gzip
import glob
import zipfile


def gunzip_file(input_path, output_path):
    """
    Decompresses a .gz file.

    Args:
        input_path (str): Path to the input .gz file.
        output_path (str): Path to save the decompressed output.
    """
    try:
        with gzip.open(input_path, 'rb') as f_in:
            with open(output_path, 'wb') as f_out:
                f_out.write(f_in.read())
    except FileNotFoundError:
         print(f"Error: Input file '{input_path}' not found.")
    except gzip.BadGzipFile:
        print(f"Error: '{input_path}' is not a valid gzip file.")
    except Exception as e:
        print(f"An error occurred: {e}")


def download_file(file_dict):
    # Determine the path to the data directory
    destdir = "./classifier_data/"
    
    urlpath = file_dict["links"]["self"]
    name = file_dict["key"]
    total_length = file_dict["size"]
    
    # make an HTTP request within a context manager
    r = requests.get(urlpath, stream=True)
        
    # implement progress bar via tqdm
    with tqdm.wrapattr(r.raw, "read", total=total_length, desc="")as raw:
    
        # save the output to a file
        with open(join(destdir, name+".gz"), 'wb') as output:
            shutil.copyfileobj(raw, output)
    
    # Determine if gzipped
    encoded = "content-encoding" in list(r.headers.keys())
    r.close()

    # Unzip the file
    with zipfile.ZipFile(join(destdir, name+".gz"), 'r') as zip_ref:
        zip_ref.extractall(destdir)
    os.remove(join(destdir, name+".gz"))

    # Gunzip file
    #if encoded:
    #    gunzip_file(join(destdir, name+".gz"), join(destdir, name))
    #    os.remove(join(destdir, name+".gz"))
    #else:
    #    os.rename(join(destdir, name+".gz"), join(destdir, name))

    return None


def check_download():
    current_dir = "/opt/data/"
    files = glob.glob(os.path.join(os.path.join(current_dir,"**"),"*.parquet"), recursive=True)
    if len(files) == 0:
        return False
    else:
        return True
    

def check_MPACT_download():
    current_dir = "/opt/data/"
    files = glob.glob(os.path.join(os.path.join(current_dir,"**"),"MPACT*.pth"), recursive=True)
    if len(files) == 0:
        return False
    else:
        return True
        

def download_methyl_anno():

    is_downloaded = check_download()
    if is_downloaded:
        return

    recordID = "16580408"
    url = 'https://zenodo.org/api/records/'
    r = requests.get(url+recordID)
    js = json.loads(r.text)

    for file_dict in js["files"]:
        download_file(file_dict)


def download_MPACT():

    is_downloaded = check_MPACT_download()
    if is_downloaded:
        return

    recordID = "16581863"
    url = 'https://zenodo.org/api/records/'
    r = requests.get(url+recordID)
    js = json.loads(r.text)

    for file_dict in js["files"]:
        download_file(file_dict)


import h5py
from projectframe import ProjectFrame
from intervalframe import IntervalFrame
import ngsfragments as ngs
from ngsfragments.plot.plot_plt import plot_cnv
import os
import glob
import numpy as np

# Local imports
from ..data.import_data import get_data_file
from ..core.utilities import predict_data_type
from ..core.microarray.RawArray import RawArray
from ..core.sequencing import read_methyldackel
from ..recipes.recipes import MPACT_process_raw


def MPACT_process_single(args):
    """
    """

    # Determine data type
    check_idats = glob.glob(args.input_data + "*.idat")
    if len(check_idats) > 0:
        input_data = check_idats
    else:
        input_data = [args.input_data]
    classifications = MPACT_process_raw(input_data = input_data,
                                        impute = args.impute,
                                        regress = args.regress,
                                        probability_threshold = args.probability_threshold,
                                        max_contamination_fraction = args.max_contamination_fraction,
                                        call_cnvs = args.call_cnvs,
                                        verbose = args.verbose)
    classifications.to_csv(args.out, header=True, index=True, sep="\t")


def microarray_call_cnvs(array_info, normal_450k_references, normal_EPIC_references, args):
    """
    """

    # Read copy number values
    cn = array_info.get_cn()

    # Filter normal
    if "450k" in array_info.array_type[0]:
        cn = cn.exact_match(normal_450k_references)
        normal_references = normal_450k_references.exact_match(cn)
    elif "EPIC" in array_info.array_type[0]:
        cn = cn.exact_match(normal_EPIC_references)
        normal_references = normal_EPIC_references.exact_match(cn)
    
    # Call CNVs
    cnvs = ngs.segment.CNVcaller(genome_version = args.genome,
                                        scStates = None,
                                        n_per_bin = args.n_probes,
                                        n_per_bin_hmm = args.n_probes_hmm)
    cnvs.predict_cnvs(data = cn,
                        normal_data = normal_references)
    if args.zscore:
        cnvs.calculate_zscore()

    # Plot CNVs
    if args.plot:
        sample = cn.df.columns.values[0]
        ngs.plot.plot_cnv(cnvs.pf,
                            obs = sample,
                            show = False,
                            save = sample+"_cnvs.pdf",
                            plot_max = 2,
                            plot_min = -2)
    # Write seg file
    if args.segs:
        sample = cn.df.columns.values[0]
        # Write seg file
        if args.segs:
            seg_fn = sample + ".seg"
            ngs.segment.cnv_utilities.write_seg_file(cnvs.pf,
                                                        seg_fn,
                                                        sample)
            
    # Write annotations
    if args.anno_file:
        cnvs.pf.anno.engine.df.to_csv(sample+"_metrics.txt", header=True, index=True, sep="\t")

    # Write seg annotations
    if args.anno_segs:
        sample = cn.df.columns.values[0]
        df = cnvs.pf.obs_intervals[sample]["cnv_segments"].df
        df.loc[:,"chrom"] = cnvs.pf.obs_intervals[sample]["cnv_segments"].index.labels
        df.loc[:,"start"] = cnvs.pf.obs_intervals[sample]["cnv_segments"].index.starts
        df.loc[:,"end"] = cnvs.pf.obs_intervals[sample]["cnv_segments"].index.ends
        df.loc[:,"sample"] = sample
        df = df.loc[:,['sample', 'chrom', 'start', 'end', 'copy_number', 'event', 'subclone_status',
                    'logR_Copy_Number', 'Corrected_Copy_Number', 'Corrected_Call', 'median']]
        df.to_csv(sample+"_seg_annotations.seg", header=True, index=False, sep="\t")

        # Drop columns
        df.drop(columns=["chrom", "start", "end"], inplace=True)

    # Write bins
    if args.bins_file:
        sample = cn.df.columns.values[0]
        df = cnvs.pf.intervals["cnv_bins"].df
        df.loc[:,"chrom"] = cnvs.pf.intervals["cnv_bins"].index.labels
        df.loc[:,"start"] = cnvs.pf.intervals["cnv_bins"].index.starts
        df.loc[:,"end"] = cnvs.pf.intervals["cnv_bins"].index.ends
        df.loc[:,"ratio"] = df.loc[:,sample].values
        if args.zscore:
            dfz = cnvs.pf.intervals["cnv_zscore_bins"].df
            df.loc[:,"zscore"] = dfz.loc[:,sample].values
            df = df.loc[:,['chrom', 'start', 'end', 'ratio', 'zscore']]
        else:
            df = df.loc[:,['chrom', 'start', 'end', 'ratio']]
        df.to_csv(sample+"_bins.txt", header=True, index=False, sep="\t")

        # Drop columns
        if args.zscore:
            df.drop(columns=["chrom", "start", "end", "ratio", "zscore"], inplace=True)
        else:
            df.drop(columns=["chrom", "start", "end", "ratio"], inplace=True)

    return None


def sequencing_call_cnvs(bedgraph, args):
    """
    """

    # Read coverage data
    cov = read_methyldackel(np.array([bedgraph]), read_coverage=True)

    # Call CNVs
    cnvs = ngs.segment.CNVcaller(genome_version = args.genome,
                                        scStates = None,
                                        n_per_bin = args.n_per_bin,
                                        n_per_bin_hmm = args.n_per_bin_hmm)
    cnvs.predict_cnvs(data = cov)
    if args.zscore:
        cnvs.calculate_zscore()

    # Plot CNVs
    if args.plot:
        sample = cov.df.columns.values[0]
        ngs.plot.plot_cnv(cnvs.pf,
                        obs = sample,
                        show = False,
                        save = sample+"_cnvs.pdf",
                        plot_max = 5,
                        plot_min = -3)
    # Write seg file
    if args.segs:
        sample = cov.df.columns.values[0]
        # Write seg file
        if args.segs:
            seg_fn = sample + ".seg"
            ngs.segment.cnv_utilities.write_seg_file(cnvs.pf,
                                                        seg_fn,
                                                        sample)
            
    # Write annotations
    if args.anno_file:
        cnvs.pf.anno.engine.df.to_csv(sample+"_metrics.txt", header=True, index=True, sep="\t")

    # Write seg annotations
    if args.anno_segs:
        sample = cov.df.columns.values[0]
        df = cnvs.pf.obs_intervals[sample]["cnv_segments"].df
        df.loc[:,"chrom"] = cnvs.pf.obs_intervals[sample]["cnv_segments"].index.labels
        df.loc[:,"start"] = cnvs.pf.obs_intervals[sample]["cnv_segments"].index.starts
        df.loc[:,"end"] = cnvs.pf.obs_intervals[sample]["cnv_segments"].index.ends
        df.loc[:,"sample"] = sample
        df.loc[:,"n_bins"] = ((df.loc[:,"end"].values - df.loc[:,"start"].values) / args.bin_size).astype(int)
        df = df.loc[:,['sample', 'chrom', 'start', 'end', 'copy_number', 'event', 'subclone_status',
                    'logR_Copy_Number', 'Corrected_Copy_Number', 'Corrected_Call', 'var', 'n_bins', 'median']]
        df.to_csv(sample+"_seg_annotations.seg", header=True, index=False, sep="\t")

        # Drop columns
        df.drop(columns=["chrom", "start", "end"], inplace=True)

    # Write bins
    if args.bins_file:
        sample = cov.df.columns.values[0]
        df = cnvs.pf.intervals["cnv_bins"].df
        df.loc[:,"chrom"] = cnvs.pf.intervals["cnv_bins"].index.labels
        df.loc[:,"start"] = cnvs.pf.intervals["cnv_bins"].index.starts
        df.loc[:,"end"] = cnvs.pf.intervals["cnv_bins"].index.ends
        df.loc[:,"ratio"] = df.loc[:,sample].values
        if args.zscore:
            dfz = cnvs.pf.intervals["cnv_zscore_bins"].df
            df.loc[:,"zscore"] = dfz.loc[:,sample].values
            df = df.loc[:,['chrom', 'start', 'end', 'ratio', 'zscore']]
        else:
            df = df.loc[:,['chrom', 'start', 'end', 'ratio']]
        df.to_csv(sample+"_bins.txt", header=True, index=False, sep="\t")

        # Drop columns
        if args.zscore:
            df.drop(columns=["chrom", "start", "end", "ratio", "zscore"], inplace=True)
        else:
            df.drop(columns=["chrom", "start", "end", "ratio"], inplace=True)

    return None



    



# cython: language_level=3

cimport cython
import numpy as np
cimport numpy as np

from ailist.array_query_core cimport pointer_to_numpy_array
from ailist.LabeledIntervalArray_core cimport LabeledIntervalArray, labeled_aiarray_t, labeled_aiarray_init, labeled_aiarray_add

from libc.stdint cimport uint32_t, uint8_t, uint64_t, int64_t, uint16_t, int32_t


cdef extern from "process_methyldackel.c":
    # C is include here so that it doesn't need to be compiled externally
    pass

cdef extern from "process_methyldackel.h":
    # C is include here so that it doesn't need to be compiled externally

    ctypedef struct sparse_record_t:
        int size							# Current size
        int max_size						# Maximum size
        double *values				        # Store values
        long *indices						# Store indices

    labeled_aiarray_t *find_all_intervals(const char file_names[], int length, int str_len) nogil

    void insert_betas(const char* fn, double *betas, labeled_aiarray_t *laia) nogil

    sparse_record_t *insert_sparse_betas(const char* fn, labeled_aiarray_t *laia) nogil

    sparse_record_t *insert_sparse_coverage(const char* fn, labeled_aiarray_t *laia) nogil


cdef np.ndarray pointer_to_double_numpy_array(void *ptr, np.npy_intp size)
cdef labeled_aiarray_t *_read_methyldackel(const char *file_names, int array_length, int str_len)
cpdef _extract_sparse_betas(const char *file_name, LabeledIntervalArray laia)
cpdef _extract_sparse_covs(const char *file_name, LabeledIntervalArray laia)


import numpy as np
import pandas as pd
from projectframe import ProjectFrame
from intervalframe import IntervalFrame
from ailist import LabeledIntervalArray
import ngsfragments as ngs


# Local imports
from ..merge_regions.merge_regions import merge_less_than, merge_greater_than
from ...data.import_data import get_data_file


def get_solo_cpgs(betas: IntervalFrame):
    """
    """

    # Get island file
    file = get_data_file("CpGislands_hg38.parquet")
    islands = IntervalFrame.read_parquet(file)
    shelfs = LabeledIntervalArray()
    shelfs.add(islands.starts-4000, islands.ends+4000, islands.index.labels)
    shelfs = shelfs.merge()

    n = shelfs.nhits_from_LabeledIntervalArray(betas.index)
    solo = betas.iloc[n == 0,:]

    return solo


def hypomethylated_regions(betas: IntervalFrame,
                           column: str,
                           sort: bool = True,
                           n_cutoff: int = 5,
                           max_length: int = 500,
                           smooth: int = 0,
                           beta_cutoff: float = 0.6):
    """
    """

    # Sort
    if sort:
        iframe = betas.iloc[betas.index.sorted_index(),:]
    else:
        # Copy
        iframe = betas.copy()

    # Smooth
    if smooth > 0:
        iframe.df.loc[:,column] = iframe.df.loc[:,column].groupby(iframe.index.labels).rolling(smooth, center=True, win_type="gaussian").mean().values

    # Remove nans
    iframe = iframe.iloc[~pd.isnull(iframe.df.loc[:,column].values),:]
    
    # Merge
    merged_intervals = merge_less_than(iframe.index, iframe.df.loc[:,column].values, beta_cutoff)
    n = betas.index.nhits_from_LabeledIntervalArray(merged_intervals)
    merged_iframe = IntervalFrame(intervals=merged_intervals)
    merged_iframe.df.loc[:,"n"] = n
    merged_iframe.annotate(betas, column, method="mean")

    # Filter
    chosen = np.logical_and((merged_iframe.ends-merged_iframe.starts) < max_length, n > n_cutoff)
    merged_iframe = merged_iframe.iloc[chosen,:]

    return merged_iframe


def hypermethylated_regions(betas: IntervalFrame,
                           column: str,
                           sort: bool = True,
                           n_cutoff: int = 5,
                           max_length: int = 500,
                           smooth: int = 0,
                           beta_cutoff: float = 0.6):
    """
    """

    # Sort
    if sort:
        iframe = betas.iloc[betas.index.sorted_index(),:]
    else:
        # Copy
        iframe = betas.copy()

    # Smooth
    if smooth > 0:
        iframe.df.loc[:,column] = iframe.df.loc[:,column].groupby(iframe.index.labels).rolling(smooth, center=True, win_type="gaussian").mean().values

    # Remove nans
    iframe = iframe.iloc[~pd.isnull(iframe.df.loc[:,column].values),:]
    
    # Merge
    merged_intervals = merge_greater_than(iframe.index, iframe.df.loc[:,column].values, beta_cutoff)
    n = betas.index.nhits_from_LabeledIntervalArray(merged_intervals)
    merged_iframe = IntervalFrame(intervals=merged_intervals)
    merged_iframe.df.loc[:,"n"] = n
    merged_iframe.annotate(betas, column, method="mean")

    # Filter
    chosen = np.logical_and((merged_iframe.ends-merged_iframe.starts) < max_length, n > n_cutoff)
    merged_iframe = merged_iframe.iloc[chosen,:]

    return merged_iframe


def get_PMDs(betas: IntervalFrame,
            column: str,
            min_length: int = 50000,
            max_length: int = 10000000,
            beta_cutoff: int = 0.4):
    """
    """

    # Remove islands, shores, and shelves
    solo = get_solo_cpgs(betas)
    #mean_methyl = np.mean(betas.df.loc[:,column].values)

    # Determine pmds
    pmds = hypomethylated_regions(solo,
                                  column = column,
                                  sort=True,
                                  smooth=101,
                                  max_length = max_length,
                                  beta_cutoff = beta_cutoff)
    
    # Filter smaller regions
    pmds = pmds.iloc[(pmds.ends - pmds.starts) > min_length,:]

    return pmds


def get_LMRs(betas: IntervalFrame,
            column: str,
            n_cutoff: int = 5,
            min_length: int = 100,
            max_length: int = 1000,
            beta_cutoff: int = 0.4):
    """
    """

    # Determine LMRs
    lmrs = hypomethylated_regions(betas,
                                  column = column,
                                  sort = True,
                                  smooth = 3,
                                  n_cutoff = n_cutoff,
                                  max_length = max_length,
                                  beta_cutoff = beta_cutoff)
    
    # Filter UMRs
    lmrs = lmrs.iloc[lmrs.df.loc[:,"mean"].values > 0.1,:]
    
    # Filter smaller regions
    lmrs = lmrs.iloc[(lmrs.ends - lmrs.starts) > min_length,:]

    return lmrs


def hypermethylated_binary_regions(pf: ProjectFrame,
                            obs: str,
                            key: str = "binary_betas",
                            n_cutoff: int = 5,
                            verbose: bool = False) -> ProjectFrame:
    """
    """

    # Get signal
    signal_file = pf.intervals[key].loc[:,[obs]]

    # Iterate over chromosomes
    chroms = signal_file.index.unique_labels
    total_peaks = LabeledIntervalArray()
    for chrom in chroms:
        if verbose: print(chrom, flush=True)
        chrom_signal = signal_file.loc[chrom,:]
        position = chrom_signal.starts
        values = np.zeros(position[-1] + 1)
        values[position] = chrom_signal.df.loc[:,obs].values
        values = pd.Series(values, index = np.arange(values.shape[0]))
        #values = signal_file.loc[chrom,:].df.loc[:,obs].astype(float)
        #values.index = np.arange(values.shape[0])
        if pd.isnull(values).sum() > 0:
            continue
        peaks = ngs.peak_calling.CallPeaks.call_peaks(values, str(chrom), min_length=10, max_length=100)

        if peaks.size > 0:
            # Re-index segments
            #peaks.index_with_aiarray(signal_file.loc[chrom,:].index)
            total_peaks.append(peaks)
            #new_peaks = LabeledIntervalArray()
            #new_peaks.add(signal_file.starts[peaks.starts], signal_file.ends[peaks.ends], np.repeat(chrom, peaks.size))
            #total_peaks.append(new_peaks)

    # Create IntervalFrame
    hypermethylated_regions = IntervalFrame(total_peaks)
    hypermethylated_regions.annotate(signal_file, column=obs, method="mean")
    hypermethylated_regions.annotate(signal_file, column=obs, method="n")

    # Filter
    hypermethylated_regions = hypermethylated_regions.iloc[hypermethylated_regions.df.loc[:,"n"].values > n_cutoff,:]

    # Assign to projectframe
    pf.add_obs_intervals(obs, "hypermethylated_binary_regions", hypermethylated_regions)
    
    return pf


def hypomethylated_binary_regions(pf: ProjectFrame,
                            obs: str,
                            key: str = "binary_betas",
                            n_cutoff: int = 5,
                            mean_cutoff: float = -0.5,
                            min_length: int = 5,
                            max_length: int = 1000,
                            gap: int = 25,
                            verbose: bool = False) -> ProjectFrame:
    """
    """

    # Get signal
    signal_file = pf.intervals[key].loc[:,[obs]]

    # Iterate over chromosomes
    chroms = signal_file.index.unique_labels
    total_peaks = LabeledIntervalArray()
    for chrom in chroms:
        if verbose: print(chrom, flush=True)
        chrom_signal = signal_file.loc[chrom,:]
        position = chrom_signal.starts
        values = np.zeros(position[-1] + 1)
        values[position] = chrom_signal.df.loc[:,obs].values * -1
        values = pd.Series(values, index = np.arange(values.shape[0]))
        #values = signal_file.loc[chrom,:].df.loc[:,obs].astype(float)
        #values.index = np.arange(values.shape[0])
        #values = values * -1
        if pd.isnull(values).sum() > 0:
            continue
        peaks = ngs.peak_calling.CallPeaks.call_peaks(values, str(chrom), min_length=min_length, max_length=max_length, merge_distance=gap)

        if peaks.size > 0:
            # Re-index segments
            #peaks.index_with_aiarray(signal_file.loc[chrom,:].index)
            peaks = peaks.merge(gap)
            total_peaks.append(peaks)
            #new_peaks = LabeledIntervalArray()
            #new_peaks.add(signal_file.starts[peaks.starts], signal_file.ends[peaks.ends], np.repeat(chrom, peaks.size))
            #total_peaks.append(new_peaks)

    # Create IntervalFrame
    hypomethylated_regions = IntervalFrame(total_peaks)
    hypomethylated_regions.annotate(signal_file, column=obs, method="mean")
    hypomethylated_regions.annotate(signal_file, column=obs, method="n")

    # Filter
    hypomethylated_regions = hypomethylated_regions.iloc[hypomethylated_regions.df.loc[:,"n"].values > n_cutoff,:]
    hypomethylated_regions = hypomethylated_regions.iloc[hypomethylated_regions.df.loc[:,"mean"].values < mean_cutoff,:]

    # Assign to projectframe
    pf.add_obs_intervals(obs, "hypomethylated_binary_regions", hypomethylated_regions)
    
    return pf


def bin_data(signal_iframe: IntervalFrame,
             chromosome: str,
             key: str = "betas",
             genome_version : str = "hg38",
             bin_size: int = 10,
             inverse: bool = False) -> np.ndarray:
    """
    Bin data.

    Parameters
    ----------
        signal_iframe : IntervalFrame
            IntervalFrame of signal.
        chromosome : str
            Chromosome to bin.
        key : str
            Key to bin.
        genome_version : str
            Genome version.
        bin_size : int
            Bin size.
    
    Returns
    -------
        V : np.ndarray
            Binned signal.
    """

    # Get chromosome length
    import genome_info
    genome = genome_info.GenomeInfo(genome_version)
    try:
        chrom_length = genome["chrom_sizes"][chromosome]
    except KeyError:
        return None
    bin_results = IntervalFrame.from_dict_range({chromosome:chrom_length}, bin_size=bin_size)
    bin_scores = np.zeros(round(int(chrom_length) / bin_size) + 1)
    n = np.zeros(round(int(chrom_length) / bin_size) + 1)

    # Get signal
    signal_iframe = signal_iframe.loc[chromosome,:]
    starts = signal_iframe.starts
    ends = signal_iframe.ends
    if inverse:
        scores = signal_iframe.df.loc[:,key].values * -1
    else:
        scores = signal_iframe.df.loc[:,key].values

    # Bin
    for i in range(len(starts)):
        start = int(starts[i])
        end = int(ends[i])
        start_bin = int(start / bin_size)
        end_bin = int(end / bin_size)
        bins = np.arange(start_bin, end_bin + 1)
        bin_scores[bins] += scores[i]
        n[bins] += 1

    bin_scores[n > 0] = bin_scores[n > 0] / n[n > 0]
    if bin_results.shape[0] != len(bin_scores):
        bin_results.df.loc[:,"score"] = bin_scores[:-1]
    else:
        bin_results.df.loc[:,"score"] = bin_scores
    
    return bin_results





import numpy as np
import pandas as pd
from intervalframe import IntervalFrame
from ailist import AIList, LabeledIntervalArray
import genome_info

# Local imports
from ..stats.beta_dist2 import *
from ...data.import_data import get_data_file
from ...core.utilities import bin_cpgs
    
    
def hypomethyl_segment(data, alpha=0.75, beta=0.5, threshold_mean=0.3):
    
    segments = AIList()
    n_regions = data.shape[0]
    start = 0
    n = 0
    i = 0
    j = 1
    llr_last = -np.inf
    while i < n_regions and j < n_regions:
        x = data[i:j, :].flatten()
        mean, llr = beta_model_test(x, alpha=alpha, beta=beta)
        
        if llr > llr_last and mean < threshold_mean:
            if n == 0:
                start = i
            j += 1
            llr_last = llr
            n += 1

        else:
            if n > 0:
                segments.add(start, j-1)
                
            i = j
            j += 1
            start = i
            n = 0
            llr_last = -np.inf
    if n > 0:
        segments.add(start, j-1)
    
    return segments


def hypermethyl_segment(data, alpha=0.5, beta=0.75, threshold_mean=0.7):
    
    segments = AIList()
    n_regions = data.shape[0]
    start = 0
    n = 0
    i = 0
    j = 1
    llr_last = -np.inf
    while i < n_regions and j < n_regions:
        x = data[i:j, :].flatten()
        mean, llr = beta_model_test(x, alpha=alpha, beta=beta)
        
        if llr > llr_last and mean > threshold_mean:
            if n == 0:
                start = i
            j += 1
            llr_last = llr
            n += 1

        else:
            if n > 0:
                segments.add(start, j-1)
                
            i = j
            j += 1
            start = i
            n = 0
            llr_last = -np.inf
    if n > 0:
        segments.add(start, j-1)
    
    return segments


def bin_fit_beta(data: IntervalFrame,
                 bins: LabeledIntervalArray = None,
                 genome_version: str = "hg38",
                 bin_size: int = 100000,
                 min_n: int = 10,
                 update: bool = False) -> IntervalFrame:
    """
    """

    # Default parameters
    x = data.df.values.flatten()
    x = x[~np.isnan(x)]
    alpha = (1-np.mean(x)) * (np.mean(x)**2) / np.var(x) - np.mean(x)
    beta = alpha * (1 / np.mean(x) - 1)

    # Get bins
    if bins is None:
        genome = genome_info.GenomeInfo(genome_version)
        bins = LabeledIntervalArray.create_bin(genome["chrom_sizes"], bin_size=bin_size)

    params = np.zeros((len(bins),2))
    for i, interval in enumerate(bins):
        methyl_betas = data.intersect(interval.start, interval.end, interval.label)
        betas = methyl_betas.df.values.flatten()
        # Remove nans
        betas = betas[~np.isnan(betas)]
        if len(betas) > min_n:
            if update:
                fitted = update_beta_prior(alpha, beta, betas)
            else:
                try:
                    fitted = estimate_beta_params(betas)
                except RuntimeError:
                    fitted = (alpha, beta)
            params[i,0] = fitted[0]
            params[i,1] = fitted[1]
        else:
            params[i,0] = alpha
            params[i,1] = beta

    # Create IntervalFrame
    params = pd.DataFrame(params, columns=["alpha", "beta"])
    iframe = IntervalFrame(df=params, intervals=bins)
        
    return iframe


def bin_beta_mean_per_sample(data: IntervalFrame,
                    bins: LabeledIntervalArray = None,
                    genome_version: str = "hg38",
                    bin_size: int = 100000,
                    min_n: int = 10,
                    update: bool = False) -> IntervalFrame:
    """
    """

    # Get samples
    samples = data.df.columns
    n_samples = len(samples)

    # Default parameters
    default_params = {}
    for sample in samples:
        x = data.df.loc[:,sample].values.flatten()
        x = x[~np.isnan(x)]
        alpha = (1-np.mean(x)) * (np.mean(x)**2) / np.var(x) - np.mean(x)
        beta = alpha * (1 / np.mean(x) - 1)
        #mean = beta_mean(alpha, beta)
        mean = np.median(np.random.beta(alpha, beta, 10000))
        default_params[sample] = (alpha, beta, mean)

    # Get bins
    if bins is None:
        genome = genome_info.GenomeInfo(genome_version)
        bins = LabeledIntervalArray.create_bin(genome["chrom_sizes"], bin_size=bin_size)

    params = np.zeros((len(bins), n_samples))
    for i, interval in enumerate(bins):
        methyl_betas = data.intersect(interval.start, interval.end, interval.label)
        
        # Get sample means
        betas = methyl_betas.df.values
        for j, sample in enumerate(samples):
            x = betas[:,j]
            # Remove nans
            x = x[~np.isnan(x)]
            if len(x) > min_n:
                if update:
                    fitted = update_beta_prior(default_params[sample][0], default_params[sample][1], x)
                else:
                    try:
                        fitted = estimate_beta_params(x)
                    except RuntimeError:
                        fitted = (default_params[sample][0], default_params[sample][1])
                #params[i,j] = beta_mean(fitted[0], fitted[1])
                params[i,j] = np.median(np.random.beta(fitted[0], fitted[1], 10000))
            else:
                params[i,j] = default_params[sample][2]

    # Create IntervalFrame
    params = pd.DataFrame(params, columns=samples)
    iframe = IntervalFrame(df=params, intervals=bins)
        
    return iframe


def methyl_blocks(data: IntervalFrame,
                  blocks: LabeledIntervalArray | IntervalFrame | str = "MethylBlocks_hg38.parquet",
                  impute_nan: bool = False,
                  method: str = "median") -> IntervalFrame:
    """
    """

    # Get blocks
    if blocks == "MethylBlocks_hg38.parquet":
        blocks = IntervalFrame.read_parquet(get_data_file("MethylBlocks_hg38.parquet"))
    elif isinstance(blocks, str):
        blocks = IntervalFrame.read_parquet(blocks)
    elif isinstance(blocks, LabeledIntervalArray):
        blocks = IntervalFrame(intervals=blocks)
    
    # Calculate block values
    overlap = data.index.intersect_from_LabeledIntervalArray(blocks.index, return_intervals=False, return_index=True)
    means = data.df.iloc[overlap[1],:].groupby(overlap[0]).mean()
    new_blocks = pd.DataFrame(np.zeros((blocks.shape[0], data.shape[1])), columns = data.df.columns)
    new_blocks[:] = np.nan
    new_blocks.iloc[means.index.values,:] = means.values
    blocks = IntervalFrame(intervals=blocks.index, df=new_blocks)

    # Impute nan
    if impute_nan:
        # Fit beta per sample
        for sample in blocks.df.columns.values:
            x = data.df.loc[:,sample].values.flatten()
            x = x[~np.isnan(x)]
            alpha = (1-np.mean(x)) * (np.mean(x)**2) / np.var(x) - np.mean(x)
            beta = alpha * (1 / np.mean(x) - 1)
            if method == "median":
                default = beta_median(alpha, beta)
            elif method == "mean":
                default = beta_mean(alpha, beta)
            else:
                default = beta_median(alpha, beta)
            
            # Impute
            blocks.df.loc[blocks.df.loc[:,sample].isna(), sample] = default

    return blocks


import pandas as pd
import numpy as np
from sklearn.svm import NuSVR
from sklearn.linear_model import HuberRegressor
#from sklearn.preprocessing import StandardScaler
#from sklearn.metrics import r2_score
from scipy.optimize import nnls
from scipy.optimize import least_squares
import h5py
from joblib import Parallel, delayed
from typing import List, Sequence, Tuple


def calculate_rmse(beta_m: pd.DataFrame,
                   ref_m: pd.DataFrame,
                   est_m: pd.DataFrame,
                   verbose: bool = False):
    """
    """

    # Iterate over samples
    rmse_v = np.zeros(beta_m.shape[0])
    for i in range(beta_m.shape[0]):
        if verbose: print(beta_m.index.values[i])
        sample_m = beta_m.iloc[i,:]
        selected = ~pd.isnull(sample_m)

        # Calculate RMSE
        reconst_m = np.matmul(ref_m.loc[:,selected].values.T, est_m.iloc[i,:].values)
        rmse_v[i] = np.sqrt(np.mean((sample_m.values[selected] - reconst_m)**2))

    return rmse_v


def rms(y, yfit):
    return np.sqrt(np.sum((y-yfit)**2))


def calculate_diff(beta_m: pd.DataFrame,
                   ref_m: pd.DataFrame,
                   est_m: pd.DataFrame,
                   verbose: bool = False):
    """
    """

    # Iterate over samples
    r2 = np.zeros(beta_m.shape[0])
    for i in range(beta_m.shape[0]):
        if verbose: print(beta_m.index.values[i])
        sample_m = beta_m.iloc[i,:]
        selected = ~pd.isnull(sample_m)

        # Calculate RMSE
        if selected.sum() > 0:
            reconst_m = np.matmul(ref_m.loc[:,selected].values.T, est_m.iloc[i,:].values)
            r2[i] = np.median(np.abs(sample_m.values[selected] - reconst_m))
        else:
            r2[i] = np.nan

    return r2


def run_cibertsort(beta_m: pd.DataFrame,
                   ref_m: pd.DataFrame,
                   n_features: int = 2000,
                   verbose: bool = False):
    """
    Parameters
    ----------
        beta_m : pd.DataFrame
            Beta values
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
        verbose : bool
            If True, print progress
    
    Returns
    -------
        est_m : pd.DataFrame
            Estimated cell fractions
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Iterate over samples
    est_m = np.zeros((beta_m.shape[0], ref_m.shape[0]))
    for i in range(beta_m.shape[0]):
        if verbose: print(beta_m.index.values[i])
        sample_m = beta_m.iloc[i,:]

        est = cibersort(sample_m, ref_m, n_features=n_features)
        est_m[i,:] = est

    # Format DataFrame
    est_m = pd.DataFrame(est_m, index=beta_m.index.values, columns=ref_m.index.values)

    return est_m


def cibersort(sample_m: np.ndarray,
              ref_m: np.ndarray,
              n_features: int = 2000,
              nu_v: List[int] = [0.25, 0.5, 0.75]) -> np.ndarray:
    """
    Parameters
    ----------
        sample_m : np.ndarray
            Sample matrix
        ref_m : np.ndarray
            Reference matrix
        n_features : int
            Number of features to use
        nu_v : List[int]
            List of nu values to use

    Returns
    -------
        est : np.ndarray
            Estimated cell fractions
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Select sample features
    sample_var = var.loc[ref_m.columns.values[~pd.isnull(sample_m.values)]]
    n_used_features = min(n_features, len(sample_var))
    selected = sample_var.index.values[np.argsort(sample_var.values)[-n_used_features:]]

    if len(selected) > 5:
        # Scale values
        sample_scaled = (sample_m.loc[selected].values - sample_m.loc[selected].values.mean()) / sample_m.loc[selected].values.std()
        ref_m_scaled = (ref_m.loc[:,selected].values - ref_m.loc[:,selected].values.mean()) / ref_m.loc[:,selected].values.std()

        est_lm = {}
        nui = 0

        # Iterate over nu values
        for nu in nu_v:
            
            # Fit model
            nusvr = NuSVR(C=1.0, nu=nu, kernel="linear")
            nusvr.fit(ref_m_scaled.T, sample_scaled)

            # Calculate coefficients
            coef_v = np.matmul(nusvr.dual_coef_, nusvr.support_vectors_)
            coef_v[coef_v<0] = 0
            total = np.sum(coef_v)
            coef_v = coef_v / total

            est_lm[nui] = coef_v
            nui += 1
        
        # Determin best nu
        rmse_m = np.zeros(len(nu_v))
        for nui in range(len(nu_v)):
            reconst_m = np.matmul(ref_m.loc[:,selected].values.T, est_lm[nui].T).T
            rmse_m[nui] = np.sqrt(np.mean((sample_m.loc[selected].values - reconst_m)**2, axis=1))
        nu_idx = np.argmin(rmse_m)

        est = est_lm[nu_idx][0]

    else:
        est = np.zeros(ref_m.shape[0])

    return est


def huber(sample_m: pd.DataFrame,
         ref_m: pd.DataFrame,
         n_features: int = 2000,
         return_rms: bool = False) -> np.ndarray:
    """
    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
        
    Returns
    -------
        coef_v : np.ndarray
            Estimated cell fractions
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Select sample features
    sample_var = var.loc[ref_m.columns.values[~pd.isnull(sample_m.values)]]
    n_used_features = min(n_features, len(sample_var))
    selected = sample_var.index.values[np.argsort(sample_var.values)[-n_used_features:]]

    if len(selected) > 5:
        # Fit model
        hr = HuberRegressor(max_iter=10000)
        hr.fit(ref_m.loc[:,selected].values.T, sample_m.loc[selected].values)

        # Calculate coefficients
        coef_v = hr.coef_
        coef_v[coef_v < 0] = 0
        total = np.sum(coef_v)
        coef_v = coef_v / total

    else:
        coef_v = np.zeros(ref_m.shape[0])

    return coef_v


def huber_weighted(sample_m: pd.DataFrame,
                ref_m: pd.DataFrame,
                n_features: int = 2000,
                return_rms: bool = False) -> np.ndarray:
    """
    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
        
    Returns
    -------
        coef_v : np.ndarray
            Estimated cell fractions
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Select sample features
    sample_var = var.loc[ref_m.columns.values[~pd.isnull(sample_m.values)]]
    n_used_features = min(n_features, len(sample_var))
    selected = sample_var.index.values[np.argsort(sample_var.values)[-n_used_features:]]
    weights = var[selected].values
    weights = weights / weights.sum()

    if len(selected) > 5:
        # Fit model
        hr = HuberRegressor(max_iter=10000)
        hr.fit(np.sqrt(weights)[:, None] * ref_m.loc[:,selected].values.T,
                      np.sqrt(weights) * sample_m.loc[selected].values)

        # Calculate coefficients
        coef_v = hr.coef_
        coef_v[coef_v < 0] = 0
        total = np.sum(coef_v)
        coef_v = coef_v / total

    else:
        coef_v = np.zeros(ref_m.shape[0])

    return coef_v


def normalize_coefs(coef_v: np.ndarray) -> np.ndarray:
    """

    Parameters
    ----------
        coef_v : np.ndarray
            Coefficient vector
    
    Returns
    -------
        coef_v : np.ndarray
            Normalized coefficient vector
    """

    coef_v[coef_v < 0] = 0
    total = np.sum(coef_v)
    coef_v = coef_v / total

    return coef_v


def calculate_rmse(sample_m: pd.DataFrame,
                   ref_m: pd.DataFrame,
                   coef_v: np.ndarray) -> float:
    """

    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        coef_v : np.ndarray
            Coefficient vector
    
    Returns
    -------
        rmse : float
            Root mean squared error
    """

    reconst_m = np.matmul(ref_m, coef_v)
    rmse = np.sqrt(np.mean((sample_m - reconst_m)**2))

    return rmse


def huber_regress(ref_m: np.ndarray,
                  sample_m: np.ndarray,
                  weighted: bool = False) -> np.ndarray:
    """

    Parameters
    ----------
        ref_m : np.ndarray
            Reference matrix
        sample_m : np.ndarray
            Sample matrix
        weighted : bool
            If True, use weighted regression
    
    Returns
    -------
        rmse : float
    """

    # Fit model
    hr = HuberRegressor(max_iter=10000)
    if weighted:
        weights = ref_m.var(axis=1)
        weights = weights / weights.sum()
        hr.fit(np.sqrt(weights)[:, None] * ref_m,
                        np.sqrt(weights) * sample_m)
        score = hr.score(ref_m, sample_m, sample_weight=weights)
    else:
        hr.fit(ref_m, sample_m)
        score = hr.score(ref_m, sample_m)

    # Calculate coefficients
    if np.sum(hr.coef_ > 0) == 0:
        #rmse = np.nan
        coef_v = np.zeros(ref_m.shape[1])
    else:
        coef_v = normalize_coefs(hr.coef_)
        #rmse = calculate_rmse(sample_m, ref_m, coef_v)

    return coef_v, score


def nnls_regress(ref_m: np.ndarray,
                  sample_m: np.ndarray,
                  weighted: bool = False) -> Tuple[float,float]:
    """

    Parameters
    ----------
        ref_m : np.ndarray
            Reference matrix
        sample_m : np.ndarray
            Sample matrix
        weighted : bool
            If True, use weighted regression
    
    Returns
    -------
        rmse : float
    """

    # Fit model
    if weighted:
        weights = ref_m.var(axis=1)
        weights = weights / weights.sum()
        coef_v = nnls(np.sqrt(weights)[:, None] * ref_m,
                        np.sqrt(weights) * sample_m)[0]
    else:
        coef_v = nnls(ref_m, sample_m)[0]

    # Calculate coefficients
    if np.sum(coef_v > 0) == 0:
        rmse = np.nan
        coef_v = np.zeros(ref_m.shape[1])
    else:
        coef_v = normalize_coefs(coef_v)
        rmse = calculate_rmse(sample_m, ref_m, coef_v)

    return coef_v, rmse


def permute_regress(ref_m: np.ndarray,
                    sample_m: np.ndarray,
                    weighted: bool = False,
                    method: str = "huber") -> np.ndarray:
    """

    Parameters
    ----------
        ref_m : np.ndarray
            Reference matrix
        sample_m : np.ndarray
            Sample matrix
        weighted : bool
            If True, use weighted regression
    
    Returns
    -------
        rmse : float
    """

    # Selecte method
    if method == "huber":
        func = huber_regress
    elif method == "nnls":
        func = nnls_regress

    # Permute
    perm = np.random.permutation(sample_m)

    # Run regression
    coefs, r = func(ref_m, perm, weighted)

    return r


def permute(sample_m: pd.DataFrame,
            ref_m: pd.DataFrame,
            n_features: int = 2000,
            n: int = 1000,
            n_jobs: int = 1,
            method: str = "huber",
            weighted: bool = False) -> np.ndarray:
    """
    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
        
    Returns
    -------
        coef_v : np.ndarray
            Estimated cell fractions
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Select sample features
    sample_var = var.loc[ref_m.columns.values[~pd.isnull(sample_m.values)]]
    n_used_features = min(n_features, len(sample_var))
    selected = sample_var.index.values[np.argsort(sample_var.values)[-n_used_features:]]

    # Determine regression function
    if method == "huber":
        func = huber_regress
    elif method == "nnls":
        func = nnls_regress
    

    if len(selected) > 5:
        obs_coefs, obs_r = func(ref_m.loc[:,selected].values.T, sample_m.loc[selected].values, weighted)
        rmse = Parallel(n_jobs=n_jobs)(delayed(permute_regress)(ref_m.loc[:,selected].values.T, sample_m.loc[selected].values, weighted, method) for i in range(n))
        rmse = np.array(rmse)

    else:
        rmse = np.zeros(n)
        rmse[:] = np.nan

    # Calculate pvalue
    #pvalue = np.sum(rmse < obs_r) / n
    pvalue = np.sum(rmse > obs_r) / n
    pvalue = max(pvalue, 1/n)

    return obs_coefs, obs_r, rmse, pvalue


def NNLS(sample_m: pd.DataFrame,
         ref_m: pd.DataFrame,
         n_features: int = 2000) -> np.ndarray:
    """

    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
    
    Returns
    -------
        coef_v : np.ndarray
            Estimated cell fractions
    
    
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Select sample features
    sample_var = var.loc[ref_m.columns.values[~pd.isnull(sample_m.values)]]
    n_used_features = min(n_features, len(sample_var))
    selected = sample_var.index.values[np.argsort(sample_var.values)[-n_used_features:]]

    if len(selected) > 5:
        # Fit model
        coef_v = nnls(ref_m.loc[:,selected].values.T, sample_m.loc[selected].values)[0]

        # Calculate coefficients
        coef_v[coef_v < 0] = 0
        total = np.sum(coef_v)
        coef_v = coef_v / total

    else:
        coef_v = np.zeros(ref_m.shape[0])

    return coef_v


def NNLS_weighted(sample_m: pd.DataFrame,
                  ref_m: pd.DataFrame,
                  n_features: int = 2000) -> np.ndarray:
    """

    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
    
    Returns
    -------
        coef_v : np.ndarray
            Estimated cell fractions
    
    
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Select sample features
    sample_var = var.loc[ref_m.columns.values[~pd.isnull(sample_m.values)]]
    n_used_features = min(n_features, len(sample_var))
    selected = sample_var.index.values[np.argsort(sample_var.values)[-n_used_features:]]
    weights = var[selected].values
    weights = weights / weights.sum()

    if len(selected) > 5:
        # Fit model
        coef_v = nnls(np.sqrt(weights)[:, None] * ref_m.loc[:,selected].values.T,
                      np.sqrt(weights) * sample_m.loc[selected].values)[0]

        # Calculate coefficients
        coef_v[coef_v < 0] = 0
        total = np.sum(coef_v)
        coef_v = coef_v / total

    else:
        coef_v = np.zeros(ref_m.shape[0])

    return coef_v


def least_square(sample_m: pd.DataFrame,
                ref_m: pd.DataFrame,
                n_features: int = 2000) -> np.ndarray:
    """

    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
        
    Returns
    -------
        coef_v : np.ndarray
            Estimated cell fractions
    """

    # Calculate variable features
    var = ref_m.var(axis=0)

    # Select sample features
    sample_var = var.loc[ref_m.columns.values[~pd.isnull(sample_m.values)]]
    n_used_features = min(n_features, len(sample_var))
    selected = sample_var.index.values[np.argsort(sample_var.values)[-n_used_features:]]

    if len(selected) > 10:
        # Fit model
        weights = np.zeros(ref_m.shape[0])
        obj_func = lambda weights: np.linalg.norm(sample_m.loc[selected].values - np.dot(weights,ref_m.loc[:,selected].values))
        coef_v = least_squares(obj_func, weights).x

        # Calculate coefficients
        coef_v[coef_v < 0] = 0
        total = np.sum(coef_v)
        coef_v = coef_v / total

    else:
        coef_v = np.zeros(ref_m.shape[0])

    return coef_v


def sample_iter_decon(sample_m: pd.DataFrame,
                        ref_m: pd.DataFrame,
                        n_features: int = 2000,
                        method: str = "huber") -> pd.Series:
    """

    Parameters
    ----------
        sample_m : pd.DataFrame
            Sample matrix
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
        method : str
            Method to use for deconvolution
    
    Returns
    -------
        est_m : pd.Series
            Estimated cell fractions
    """

    est_m = pd.Series(np.zeros(ref_m.shape[0]),
                      index = ref_m.index.values)

    select = np.arange(ref_m.shape[0])
    n_select = 0

    while len(select) != n_select:
        n_select = len(select)
        if n_select == 0:
            break
        
        if method == "huber":
            coef_v = huber(sample_m,
                            ref_m.iloc[select,:],
                            n_features=n_features)
        elif method == "ciber":
            coef_v = cibersort(sample_m,
                                ref_m.iloc[select,:],
                                n_features=n_features)
        elif method == "nnls":
            coef_v = NNLS(sample_m,
                            ref_m.iloc[select,:],
                            n_features=n_features)
        elif method == "wnnls":
            coef_v = NNLS_weighted(sample_m,
                                   ref_m.iloc[select,:],
                                   n_features=n_features)
        elif method == "lstsq":
                coef_v = least_square(sample_m,
                              ref_m.iloc[select,:],
                              n_features=n_features)
        select = select[coef_v > 0.0]

    if len(select) > 0:
        est_m.values[select] = coef_v

    return est_m


def iter_decon(beta_m: pd.DataFrame,
               ref_m: pd.DataFrame,
               n_features: int = 2000,
               method: str = "huber",
               controls: Sequence[str] = np.array(["CONTR_INFLAM",
                                              "CONTR_REACT",
                                              "CONTR_CEBM",
                                              "PLASMA"]),
               verbose: bool = False) -> pd.DataFrame:
    """

    Parameters
    ----------
        beta_m : pd.DataFrame
            Beta values
        ref_m : pd.DataFrame
            Reference matrix
        n_features : int
            Number of features to use
        method : str
            Method to use for deconvolution
        controls : Sequence[str]
            List of control cell types
        verbose : bool
            If True, print progress

    Returns
    -------
        est_m : pd.DataFrame
            Estimated cell fractions
    """

    # Select controls
    is_control = np.in1d(ref_m.index.values, controls)

    # Initialize estimation matrix
    est_m = pd.DataFrame(np.zeros((beta_m.shape[0], ref_m.shape[0])),
                         index = beta_m.index.values,
                         columns = ref_m.index.values)
                     

    # Iterate over samples
    for i, s in enumerate(beta_m.index.values):
        if verbose: print(s)

        # Calculate first estimates
        sample_decon = sample_iter_decon(beta_m.iloc[i,:],
                                            ref_m,
                                            n_features=n_features,
                                            method=method)
        
        if np.max(sample_decon.values[~is_control]) > 0:
            top_hit = sample_decon.index.values[~is_control][np.argmax(sample_decon.values[~is_control])]
            selected = np.append(controls, top_hit)

            # Calculate estimates
            sample_decon = sample_iter_decon(beta_m.iloc[i,:],
                                                ref_m.loc[selected,:],
                                                n_features=n_features,
                                                method=method)
        # Set estimates
        est_m.loc[s, sample_decon.index.values] = sample_decon.values
        
    return est_m


def make_signature_matrix(ref_m: pd.DataFrame,
                          groups: Sequence[str],
                          method: str = "mean") -> pd.DataFrame:
    """

    Parameters
    ----------
        ref_m : pd.DataFrame
            Reference matrix
        groups : Sequence[str]
            List of groups
        method : str
            Method to use for calculating signature matrix

    Returns
    -------
        sig_matrix : pd.DataFrame
            Signature matrix
    """

    # Calculate matrix
    if method == "mean":
        sig_matrix = ref_m.groupby(groups).mean()
    elif method == "median":
        sig_matrix = ref_m.groupby(groups).median()
    else:
        raise ValueError("Method not implemented")
    
    return sig_matrix


from intervalframe import IntervalFrame
import ngsfragments as ngs

# Local imports
from ...data.import_data import get_data_file
from ...core.microarray.RawArray import RawArray


def microarray_call_cnvs(array_info: RawArray,
                         genome_version: str = "hg38",
                         n_probes: int = 10,
                         n_probes_hmm: int = 15,
                         zscore: bool = False,
                         n_mads: float = 1.0) -> ngs.segment.CNVcaller:
    """
    """

    # Read copy number values
    cn = array_info.get_cn()

    # Get normal references
    if genome_version  == "hg19":
        normal_450k_references = IntervalFrame.read_parquet(get_data_file("normal450k_hg19_cn.parquet"))
        normal_EPIC_references = IntervalFrame.read_parquet(get_data_file("normalEPIC_hg19_cn.parquet"))
    elif genome_version == "hg38":
        normal_450k_references = IntervalFrame.read_parquet(get_data_file("normal450k_hg38_cn.parquet"))
        normal_EPIC_references = IntervalFrame.read_parquet(get_data_file("normalEPIC_hg38_cn.parquet"))

    # Filter normal
    if "450k" in array_info.array_type[0]:
        cn = cn.exact_match(normal_450k_references)
        normal_references = normal_450k_references.exact_match(cn)
    elif "EPIC" in array_info.array_type[0]:
        cn = cn.exact_match(normal_EPIC_references)
        normal_references = normal_EPIC_references.exact_match(cn)
    
    # Call CNVs
    cnvs = ngs.segment.CNVcaller(genome_version = genome_version,
                                        scStates = None,
                                        n_per_bin = n_probes,
                                        n_per_bin_hmm = n_probes_hmm,
                                        n_mads = n_mads)
    cnvs.predict_cnvs(data = cn,
                        normal_data = normal_references)
    if zscore:
        cnvs.calculate_zscore()

    return cnvs


import glob
import os
from typing import List
from projectframe import ProjectFrame
import gc
import numpy as np
import pandas as pd
from ailist import LabeledIntervalArray
from intervalframe import IntervalFrame

# Local imports
from ..data.import_data import get_data_file
from .sequencing import read_methyldackel
from .microarray import RawArray


def bams_in_directory(directory: str) -> bool:
    """
    Check if bams are present in directory
    """

    files = glob.glob(os.path.join(directory, "*.bam"))
    if len(files) > 0:
        return True
    else:
        return False


def idats_in_directory(directory: str) -> bool:
    """
    Check if idats are present in directory
    """

    files = glob.glob(os.path.join(directory, "*.idat"))
    if len(files) > 0:
        return True
    else:
        return False


def read_methylbed(filenames: str,
                   read_cov: bool = False,
                   verbose: bool = False) -> pd.DataFrame:
    """
    """

    if verbose:
        print("Reading", filenames, flush=True)

    # Initialize
    filename = filenames[0]
    # Read
    x = pd.read_csv(filename, header=None, index_col=None, sep='\s+')

    # Filter
    if read_cov:
        x.loc[:,filename] = x.iloc[:,9].values
    else:
        x.loc[:,filename] = x.iloc[:,10].values / 100.0
    x = x.loc[x.iloc[:,3].values == "m",:]
    ail = LabeledIntervalArray()
    ail.add(x.iloc[:,1].values, x.iloc[:,2].values + 1, x.iloc[:,0].values)
    x = IntervalFrame(df=x.drop(columns=x.columns.values[x.columns.values!=filename]), intervals=ail)

    # Create list
    iframes = []
    iframes.append(x)

    for filename in filenames[1:]:
        # Read
        smp = pd.read_csv(filename, header=None, index_col=None, sep="\s+")

        # Filter
        if read_cov:
            smp.loc[:,filename] = smp.iloc[:,9].values
        else:
            smp.loc[:,filename] = smp.iloc[:,10].values / 100.0
        smp = smp.loc[smp.iloc[:,3].values == "m",:]
        ail = LabeledIntervalArray()
        ail.add(smp.iloc[:,1].values, smp.iloc[:,2].values + 1, smp.iloc[:,0].values)
        smp = IntervalFrame(df=smp.drop(columns=smp.columns.values[smp.columns.values!=filename]), intervals=ail)
        iframes.append(smp)

    # Merge
    iframe = IntervalFrame.combine(iframes)
    del iframes
    gc.collect()

    return iframe


def read_bed(filenames: str,
             read_cov: bool = False,
             verbose: bool = False) -> pd.DataFrame:
    """
    """

    if verbose:
        print("Reading", filenames, flush=True)

    # Initialize
    x = IntervalFrame()

    for filename in filenames:
        # Read
        x = IntervalFrame.read_bed(filename, header=None, skipfirst=True)

        # Filter
        if read_cov:
            x.df.loc[:,filename] = x.df.iloc[:,1].values + x.df.iloc[:,2].values
        else:
            x.df.loc[:,filename] = x.df.iloc[:,0].values / 100.0
        x.df = x.df.drop(columns=x.df.columns.values[x.df.columns.values!=filename])

    return x


def read_sequencing_methylation(file_path: str,
                                read_cov: bool = False,
                                verbose: bool = False) -> IntervalFrame:
    """
    """

    # Initialize
    methyl_data = []
    methylation = None
    
    # Read uncompressed bedgraph files
    ## Check if file_path is a file name
    if os.path.isfile(file_path) and ".bedGraph" in file_path:
        seq_files = np.array([file_path])
    elif os.path.isfile(file_path) and ".bedgraph" in file_path:
        seq_files = np.array([file_path])
    else:
        seq_files = glob.glob(file_path + "/*.bedGraph*")
        seq_files = np.array(glob.glob(file_path + "/*.bedgraph*") + seq_files)
    if len(seq_files) > 0:
        if verbose:
            print("Reading", seq_files, flush=True)
        seq_methylation = read_methyldackel(seq_files, use_sparse=False, read_coverage=read_cov)
        methyl_data.append(seq_methylation)
        del seq_methylation
        gc.collect()

    # Read bed files from methylbed
    if os.path.isfile(file_path) and file_path.endswith(".bed"):
        seq_files = [file_path]
    elif os.path.isfile(file_path) and file_path.endswith(".bed.gz"):
        seq_files = [file_path]
    elif os.path.isfile(file_path) and ".bedmethyl" in file_path:
        seq_files = [file_path]
    else:
        seq_files = glob.glob(file_path + "/*.bed")
        seq_files = glob.glob(file_path + "/*.bed.gz") + seq_files
        seq_files = glob.glob(file_path + "/*.bedmethyl") + seq_files
        seq_files = np.array(glob.glob(file_path + "/*.bedmethyl.gz") + seq_files)
    if len(seq_files) > 0:
        if verbose:
            print("Reading", seq_files, flush=True)
        seq_methylation = read_methylbed(seq_files, read_cov=read_cov)
        methyl_data.append(seq_methylation)
        del seq_methylation
        gc.collect()

    # Merge
    if len(methyl_data) > 0:
        methylation = IntervalFrame.combine(methyl_data)
        del methyl_data
        gc.collect()

    return methylation


def read_methylation(file_path: str | List[str],
                     genome_version: str = "hg38",
                     verbose: bool = False) -> IntervalFrame:
    """
    """

    # Read parquet
    if file_path.endswith(".parquet"):
        methylation = IntervalFrame.read_parquet(file_path)

    else:
        # Read sequencing data
        seq_methylation = read_sequencing_methylation(file_path, verbose=verbose)

        # Read idat data
        idat_files = glob.glob(file_path + "/*.idat")
        if len(idat_files) > 0:
            if verbose:
                print("Reading", idat_files, flush=True)
            # Read
            data = RawArray(file_path, genome_version=genome_version, n_jobs=1, verbose=False)
            idat_methylation = data.get_betas(return_intervals=True)

        # Merge
        if seq_methylation is not None and len(idat_files) > 0:
            methylation = IntervalFrame.combine([seq_methylation, idat_methylation])
        elif seq_methylation is not None > 0:
            methylation = seq_methylation
        elif len(idat_files) > 0:
            methylation = idat_methylation
        else:
            raise ValueError("No methylation files found in directory")
    
    return methylation







import numpy as np
import pandas as pd
from projectframe import ProjectFrame
from intervalframe import IntervalFrame
from ailist import LabeledIntervalArray
from statsmodels.stats.proportion import proportions_ztest
from statsmodels.stats.multitest import fdrcorrection
from typing import List

# Local imports
from ...core.utilities import mean_methyl


def merge_peaks(peaks1: IntervalFrame,
                    peaks2: IntervalFrame,
                    gap: int = 1) -> IntervalFrame:
    """
    """

    new_peaks = LabeledIntervalArray()
    new_peaks.append(peaks1.index)
    new_peaks.append(peaks2.index)
    new_peaks = new_peaks.merge(gap)

    new_intervals = IntervalFrame(intervals = new_peaks)
    
    return new_intervals



def standardize_peaks(peaks: IntervalFrame,
                      standard_size: int) -> IntervalFrame:
    """
    """

    new_peaks = LabeledIntervalArray()
    half_size = int(standard_size / 2)
    midpoints = (peaks.starts + ((peaks.ends - peaks.starts) / 2)).astype(int)
    new_peaks.add(midpoints - half_size, midpoints + half_size, peaks.index.labels)
    new_peaks = new_peaks.merge()

    peaks = IntervalFrame(intervals = new_peaks)

    return peaks


def concensus_peaks(peaks: List[IntervalFrame],
                    gap: int = 1,
                    standard_size: int = None):
    """
    """

    # Get peaks
    total_peaks = peaks[0]

    # Iterate over observations
    for i, obs_peaks in enumerate(peaks):
        if i == 0:
            continue
        total_peaks = merge_peaks(total_peaks, obs_peaks, gap)

    # Standardize
    if standard_size is not None:
        total_peaks = standardize_peaks(total_peaks, standard_size)
        total_peaks = standardize_peaks(total_peaks, standard_size)
        ## Twice to make sure
        lengths = total_peaks.ends - total_peaks.starts
        total_peaks = total_peaks.iloc[lengths == standard_size,:]

    return total_peaks


def total_concensus_peaks(pf: ProjectFrame,
                            key: str = "hypomethylated_binary_regions",
                            anno_key: str = "binary_betas",
                            min_overlap: int = 1,
                            standard_size: int = None) -> ProjectFrame:
    """
    """

    # Get peaks
    obs = list(pf.obs)
    peaks = pf.obs_intervals[obs[0]][key]

    # Iterate over observations
    for o in obs[1:]:
        obs_peaks = pf.obs_intervals[o][key]
        peaks = merge_peaks(peaks, obs_peaks, min_overlap)

    # Standardize
    if standard_size is not None:
        peaks = standardize_peaks(peaks, standard_size)
        peaks = standardize_peaks(peaks, standard_size)
        ## Twice to make sure
        lengths = peaks.ends - peaks.starts
        peaks = peaks.iloc[lengths == standard_size,:]

    # Annotate peaks
    peaks = mean_methyl(peaks, pf.intervals[anno_key])

    # Assign to projectframe
    pf.add_intervals("concensus_peaks", peaks)

    return pf


def prop_test(pf: ProjectFrame,
              intervals: IntervalFrame,
              obs1: str,
              obs2: str,
              key: str = "binary_betas",
              alpha: float = 0.05,
              min_n: int = 3) -> IntervalFrame:
    """
    """

    new_peaks = IntervalFrame(intervals = intervals.index)
    new_peaks.df.loc[:,"stat"] = 0.0
    new_peaks.df.loc[:,obs1+"_n"] = 0
    new_peaks.df.loc[:,obs2+"_n"] = 0
    new_peaks.df.loc[:,obs1+"_methylated"] = 0
    new_peaks.df.loc[:,obs2+"_methylated"] = 0

    starts = new_peaks.starts
    ends = new_peaks.ends
    chroms = new_peaks.index.labels
    pvals = np.ones(new_peaks.shape[0])
    #stats = np.zeros(new_peaks.shape[0])
    for i in range(new_peaks.shape[0]):
        start = int(starts[i])
        end = int(ends[i])
        chrom = str(chroms[i])
        count = np.sum(pf.intervals[key].intersect(start, end, chrom).df.loc[:,[obs1, obs2]].values==1, axis=0)
        nobs = np.sum(pf.intervals[key].intersect(start, end, chrom).df.loc[:,[obs1, obs2]].values!=0, axis=0)
        #count = np.array([np.sum(pf1.intervals["betas"].intersect(start, end, chrom).df.loc[:,[obs1, obs2]].values==1), np.sum(signal_file2.intersect(start, end, chrom).df.loc[:,"betas"].values==1)])
        #nobs = np.array([signal_file.intersect(start, end, chrom).shape[0], signal_file2.intersect(start, end, chrom).shape[0]])
        stat, pval = proportions_ztest(count, nobs)
        pvals[i] = pval
        new_peaks.df.iloc[i,0] = stat
        new_peaks.df.iloc[i,1] = count[0]
        new_peaks.df.iloc[i,2] = count[1]
        new_peaks.df.iloc[i,3] = nobs[0]
        new_peaks.df.iloc[i,4] = nobs[1]

    # Correct
    pvals[~pd.isnull(pvals)] = fdrcorrection(pvals[~pd.isnull(pvals)])[1]

    # Add to results
    new_peaks.df.loc[:,"fdr_pval"] = pvals

    # Filter
    new_peaks = new_peaks.iloc[new_peaks.df.loc[:,"fdr_pval"].values<alpha,:]
    new_peaks = new_peaks.iloc[new_peaks.df.loc[:,"stat"].values>0,:]
    new_peaks = new_peaks.iloc[new_peaks.df.loc[:,obs1+"_n"].values>=min_n,:]

    return new_peaks


def compare_peaks(pf: ProjectFrame,
                  obs1: str,
                  obs2: str,
                  key: str = "hypomethylated_binary_regions",
                  min_overlap: int = 1,
                  alpha: float = 0.05,
                  min_n: int = 3) -> ProjectFrame:
    """
    """

    # Get peaks
    peaks1 = pf.obs_intervals[obs1][key]
    peaks2 = pf.obs_intervals[obs2][key]

    # Compare
    new_peaks = concensus_peaks(peaks1, peaks2, min_overlap)
    new_peaks.df.loc[:,"stat"] = 0.0
    new_peaks.df.loc[:,obs1+"_n"] = 0
    new_peaks.df.loc[:,obs2+"_n"] = 0
    new_peaks.df.loc[:,obs1+"_methylated"] = 0
    new_peaks.df.loc[:,obs2+"_methylated"] = 0

    starts = new_peaks.starts
    ends = new_peaks.ends
    chroms = new_peaks.index.labels
    pvals = np.zeros(new_peaks.shape[0])
    #stats = np.zeros(new_peaks.shape[0])
    for i in range(new_peaks.shape[0]):
        start = int(starts[i])
        end = int(ends[i])
        chrom = str(chroms[i])
        count = np.sum(pf.intervals["binary_betas"].intersect(start, end, chrom).df.loc[:,[obs1, obs2]].values==1, axis=0)
        nobs = np.sum(pf.intervals["binary_betas"].intersect(start, end, chrom).df.loc[:,[obs1, obs2]].values!=0, axis=0)
        #count = np.array([np.sum(pf1.intervals["betas"].intersect(start, end, chrom).df.loc[:,[obs1, obs2]].values==1), np.sum(signal_file2.intersect(start, end, chrom).df.loc[:,"betas"].values==1)])
        #nobs = np.array([signal_file.intersect(start, end, chrom).shape[0], signal_file2.intersect(start, end, chrom).shape[0]])
        stat, pval = proportions_ztest(count, nobs)
        pvals[i] = pval
        new_peaks.df.iloc[i,0] = stat
        new_peaks.df.iloc[i,1] = count[0]
        new_peaks.df.iloc[i,2] = count[1]
        new_peaks.df.iloc[i,3] = nobs[0]
        new_peaks.df.iloc[i,4] = nobs[1]

    # Correct
    pvals[~pd.isnull(pvals)] = fdrcorrection(pvals[~pd.isnull(pvals)])[1]

    # Add to results
    new_peaks.df.loc[:,"fdr_pval"] = pvals

    # Filter
    new_peaks = new_peaks.iloc[new_peaks.df.loc[:,"fdr_pval"].values<alpha,:]
    new_peaks = new_peaks.iloc[new_peaks.df.loc[:,"stat"].values>0,:]
    new_peaks = new_peaks.iloc[new_peaks.df.loc[:,obs1+"_n"].values>=min_n,:]

    # Assign to projectframe
    pf.add_obs_intervals(obs1, obs2+"_diff_peaks", new_peaks)

    return pf


import numpy as np
from scipy.ndimage import uniform_filter1d
from scipy import signal
from intervalframe import IntervalFrame
from ailist import LabeledIntervalArray


def adaptive_window_pmds(beta_values,
                         window_sizes=[155, 255, 995],
                         threshold=(0.3, 0.7), 
                         min_segment_length=750):
    """
    Use adaptive sliding windows to directly identify PMD segments based on beta value ranges.
    This method doesn't rely on HMM pre-classification.
    
    Parameters:
    -----------
    beta_values : array-like
        Array of methylation beta values (0-1)
    window_sizes : list of int
        Different window sizes to use for multi-scale analysis
    threshold : float or tuple
        If float: threshold for classifying a region as PMD (regions with averages between
        threshold and 1-threshold are considered PMDs)
        If tuple: (low_threshold, high_threshold) for PMD range
    min_segment_length : int
        Minimum length of segments to retain
        
    Returns:
    --------
    tuple
        (pmd_segments, segment_scores)
        pmd_segments: list of tuples (start_index, end_index) for each PMD
        segment_scores: array with PMD likelihood score for each position
    """
    if isinstance(threshold, (int, float)):
        # Symmetric threshold
        low_threshold = threshold
        high_threshold = 1 - threshold
    else:
        # Asymmetric threshold provided as tuple
        low_threshold, high_threshold = threshold
    
    # Multi-scale smoothing
    smoothed_values = np.zeros((len(window_sizes), len(beta_values)))
    
    for i, window in enumerate(window_sizes):
        # Edge-preserving smoothing using median filter
        smoothed_values[i] = signal.medfilt(beta_values, window)
    
    # Combine multi-scale information (average across scales)
    combined_smooth = np.mean(smoothed_values, axis=0)
    
    # Score each position by how likely it is to be a PMD
    # 1.0 = definitely PMD, 0.0 = definitely not PMD
    segment_scores = np.zeros(len(beta_values))
    
    for i, val in enumerate(combined_smooth):
        # Calculate distance from fully methylated or unmethylated
        if val < low_threshold:
            # Unmethylated region
            segment_scores[i] = 0.0
        elif val > high_threshold:
            # Fully methylated region
            segment_scores[i] = 0.0
        else:
            # PMD region - score based on how central it is in the PMD range
            # Highest score at the center of the PMD range
            mid_point = (low_threshold + high_threshold) / 2
            distance_from_mid = abs(val - mid_point)
            max_distance = (high_threshold - low_threshold) / 2
            segment_scores[i] = 1.0 - (distance_from_mid / max_distance)
    
    # Binarize scores with threshold for segment identification
    binary_mask = segment_scores > 0.5
    
    # Find contiguous segments
    pmd_segments = []
    in_segment = False
    start_pos = 0
    
    for i in range(len(binary_mask)):
        if binary_mask[i] and not in_segment:
            # Start of a new segment
            start_pos = i
            in_segment = True
        elif not binary_mask[i] and in_segment:
            # End of a segment
            if i - start_pos >= min_segment_length:
                pmd_segments.append((start_pos, i-1))
            in_segment = False
    
    # Handle case where the last segment extends to the end
    if in_segment and len(binary_mask) - start_pos >= min_segment_length:
        pmd_segments.append((start_pos, len(binary_mask)-1))
    
    return pmd_segments, segment_scores


def refine_segment_boundaries(beta_values,
                              segments,
                              window_size=5,
                              threshold=0.5):
    """
    Refine the boundaries of identified segments by looking for sharp transitions
    in methylation levels.
    
    Parameters:
    -----------
    beta_values : array-like
        Array of methylation beta values
    segments : list of tuples
        List of (start, end) tuples defining segments
    window_size : int
        Size of window to look for transitions
    threshold : float
        Threshold for change in beta value to consider a boundary
        
    Returns:
    --------
    list
        Refined segments as (start, end) tuples
    """
    refined_segments = []
    
    # Calculate derivative of beta values to detect transitions
    beta_gradient = np.abs(np.gradient(beta_values))
    
    for start, end in segments:
        # Look for a more precise start boundary
        # Search within window_size of original boundary
        search_start = max(0, start - window_size)
        search_end = min(len(beta_values) - 1, start + window_size)
        
        # Find position with maximum gradient in the search window
        max_gradient_pos = search_start + np.argmax(beta_gradient[search_start:search_end+1])
        refined_start = max_gradient_pos
        
        # Look for a more precise end boundary
        search_start = max(0, end - window_size)
        search_end = min(len(beta_values) - 1, end + window_size)
        
        max_gradient_pos = search_start + np.argmax(beta_gradient[search_start:search_end+1])
        refined_end = max_gradient_pos
        
        refined_segments.append((refined_start, refined_end))
    
    return refined_segments